# DTFIX5: more thorough tests are needed here, this is just a starting point

from __future__ import annotations

import collections.abc as c
import copy
import re
import sys
import typing as t

import pytest

from ansible.errors import AnsibleTemplateError, AnsibleUndefinedVariable
from ansible._internal._templating._jinja_bits import is_possibly_template
from ansible._internal._templating._jinja_common import CapturedExceptionMarker, MarkerError, JinjaCallContext, Marker
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate
from ansible._internal._templating._utils import TemplateContext, LazyOptions
from ansible._internal._templating._engine import TemplateEngine, TemplateOptions
from ansible._internal._templating._lazy_containers import _AnsibleLazyTemplateMixin, _AnsibleLazyTemplateList, _AnsibleLazyTemplateDict, _LazyValue, \
    _AnsibleLazyAccessTuple, UnsupportedConstructionMethodError
from ansible.module_utils._internal._datatag import AnsibleTaggedObject

from ...module_utils.datatag.test_datatag import ExampleSingletonTag
from ...test_utils.controller.display import emits_warnings

TRUST = TrustedAsTemplate()

VALUE_TO_TEMPLATE = TRUST.tag("{{ 'hello' | default('goodbye') }}")

CONTAINER_VALUES = (
    dict(hello=VALUE_TO_TEMPLATE),
    [VALUE_TO_TEMPLATE],
)


@pytest.mark.parametrize("value", CONTAINER_VALUES, ids=[type(value).__name__ for value in CONTAINER_VALUES])
def test_container_equality(value: t.Any) -> None:
    templar = TemplateEngine()

    rendered = templar.template(value)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        # NOTE: Assertion failure helper text may be misleading, since repr() will show rendered templates, which will appear to match expected values.

        lazy = _AnsibleLazyTemplateMixin._try_create(value)

        assert lazy == lazy  # pylint: disable=comparison-with-itself

        assert lazy == rendered
        assert rendered == lazy

        assert lazy != value
        assert value != lazy


@pytest.mark.parametrize("value", CONTAINER_VALUES, ids=[type(value).__name__ for value in CONTAINER_VALUES])
def test_container_format(value: t.Any) -> None:
    templar = TemplateEngine()

    rendered = templar.template(value)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        # NOTE: Assertion failure helper text may be misleading, since repr() will show rendered templates, which will appear to match expected values.

        lazy = _AnsibleLazyTemplateMixin._try_create(value)

        assert "{0}".format(lazy) == "{0}".format(rendered)


@pytest.mark.parametrize("container_type", (
    list,
))
def test_container_contains(container_type: type) -> None:
    templar = TemplateEngine()

    # including default('goodbye') as canary for flattening to a string
    value = container_type([VALUE_TO_TEMPLATE])
    rendered = templar.template(value)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        # NOTE: Assertion failure helper text may be misleading, since repr() will show rendered templates, which will appear to match expected values.

        lazy = _AnsibleLazyTemplateMixin._try_create(value)

        for src in (lazy, rendered):
            assert 'hello' in src
            assert 'goodbye' not in src


@pytest.mark.parametrize("container_type", (
    list,
    # no need to test a tuple, it's just converted to a list
))
def test_container_comparison(container_type: type) -> None:
    templar = TemplateEngine()

    # including default('goodbye') as canary for flattening to a string
    value = container_type([VALUE_TO_TEMPLATE])
    rendered = templar.template(value)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        # NOTE: Assertion failure helper text may be misleading, since repr() will show rendered templates, which will appear to match expected values.

        lazy = _AnsibleLazyTemplateMixin._try_create(value)

        assert value > rendered
        assert not (lazy > rendered)  # pylint: disable=unnecessary-negation

        assert value >= rendered
        assert lazy >= rendered

        assert rendered < value
        assert not (rendered < lazy)  # pylint: disable=unnecessary-negation

        assert rendered <= value
        assert rendered <= lazy


def test_list_sort() -> None:
    templar = TemplateEngine()

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        lazy: list = _AnsibleLazyTemplateMixin._try_create([TRUST.tag('{{ 2 }}'), TRUST.tag('{{ 1 }}')])
        lazy.sort()

        assert lazy == [1, 2]


def test_list_index() -> None:
    templar = TemplateEngine()

    rendered = templar.template(VALUE_TO_TEMPLATE)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        lazy: list = _AnsibleLazyTemplateMixin._try_create([VALUE_TO_TEMPLATE])

        assert lazy.index(rendered) == 0


def test_list_remove() -> None:
    templar = TemplateEngine()

    rendered = templar.template(VALUE_TO_TEMPLATE)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        lazy: list = _AnsibleLazyTemplateMixin._try_create([VALUE_TO_TEMPLATE])

        assert rendered in lazy

        lazy.remove(rendered)

        assert lazy == []


def test_dict_get_with_default(template_context) -> None:
    """Ensure getting a default value does not result in templating or storage of the default."""
    my_dict = _AnsibleLazyTemplateMixin._try_create({})
    value = TrustedAsTemplate().tag("{{ 1 }}")

    my_dict['x'] = TrustedAsTemplate().tag("{{ 2 }}")

    result = my_dict.get('a', value)

    assert result is value
    assert 'a' not in my_dict


def test_dict_setdefault(template_context) -> None:
    """Ensure that setdefault templates existing values, but not defaults."""
    my_dict = _AnsibleLazyTemplateMixin._try_create(dict(invalid_template=TRUST.tag("{{ 1/0 }}"), valid_template=TRUST.tag("{{ 1 }}")))
    value_for_default = TrustedAsTemplate().tag("{{ 'default' }}")

    assert my_dict.setdefault('valid_template') == 1
    assert my_dict.setdefault('valid_template', value_for_default) == 1
    assert my_dict.setdefault('nonexistent_key', value_for_default) is value_for_default

    result = my_dict.setdefault('invalid_template', value_for_default)
    assert isinstance(result, CapturedExceptionMarker)
    assert isinstance(result._marker_captured_exception, AnsibleTemplateError)

    # repeat to ensure we didn't record any change
    result = my_dict.setdefault('invalid_template', value_for_default)
    assert isinstance(result, CapturedExceptionMarker)
    assert isinstance(result._marker_captured_exception, AnsibleTemplateError)


def test_dict_pop(template_context) -> None:
    """Ensure that pop does not template or store its default, and that templating occurs before the collection is mutated."""
    my_dict = _AnsibleLazyTemplateMixin._try_create(dict(busted_template=TRUST.tag("{{ 1 / 0 }}")))
    value_for_default = TRUST.tag("{{ 1 }}")

    result = my_dict.pop("boguskey", value_for_default)

    assert result is value_for_default
    assert "boguskey" not in my_dict

    with JinjaCallContext(accept_lazy_markers=False):
        with pytest.raises(MarkerError):
            my_dict.pop('busted_template')

    assert 'busted_template' in my_dict

    result = my_dict.pop('busted_template')

    assert isinstance(result, CapturedExceptionMarker)
    assert isinstance(result._marker_captured_exception, AnsibleTemplateError)

    assert 'busted_template' not in my_dict


def test_dict_popitem(template_context):
    """Ensure popitem respects insertion order, templating of values, and that templating occurs before the collection is mutated."""
    my_dict = _AnsibleLazyTemplateMixin._try_create(dict(
        also_valid_template=TRUST.tag("{{ 0 }}"),
        busted_template=TRUST.tag("{{ 1 / 0 }}"),
        valid_template=TRUST.tag("{{ 1 }}"),
    ))

    assert my_dict.popitem() == ('valid_template', 1)

    with JinjaCallContext(accept_lazy_markers=False):
        with pytest.raises(MarkerError):
            my_dict.popitem()

    assert 'busted_template' in my_dict

    raw_result = my_dict.popitem()

    assert isinstance(raw_result, tuple)

    key, result = raw_result

    assert key == 'busted_template'
    assert isinstance(result, CapturedExceptionMarker)
    assert isinstance(result._marker_captured_exception, AnsibleTemplateError)

    assert my_dict.popitem() == ('also_valid_template', 0)

    with pytest.raises(KeyError):
        my_dict.popitem()


@pytest.mark.parametrize("native_value, expression", (
    ({}, 'obj.popitem()'),
    ([], 'obj.pop()'),
    ({}, 'obj.pop("missing")'),
    ([], 'obj.pop(1)'),
))
def test_get_errors(template_context, native_value: object, expression: str) -> None:
    """Verify getting a value from a lazy container fails the same as the native container when the requested key/index is missing or the container is empty."""
    lazy_value = _AnsibleLazyTemplateMixin._try_create(native_value)

    with pytest.raises(Exception) as native_error:
        eval(expression, dict(obj=native_value))

    with pytest.raises(type(native_error.value), match=f'^{re.escape(str(native_error.value))}$'):
        eval(expression, dict(obj=lazy_value))


def test_dict_items_and_values() -> None:
    templar = TemplateEngine()

    value = dict(key=VALUE_TO_TEMPLATE)
    rendered = templar.template(value)

    with TemplateContext(template_value=None, templar=templar, options=TemplateOptions(), stop_on_template=False):
        lazy: dict = _AnsibleLazyTemplateMixin._try_create(value)

        assert list(lazy.items()) == list(rendered.items())
        assert list(lazy.values()) == list(rendered.values())


@pytest.mark.parametrize("template, variables, expected", [
    ("{{ (d1 | items) + (d2 | items) }}", dict(d1=dict(a=1), d2=dict(b=2)), [['a', 1], ['b', 2]]),
    ("{{ (d1 | items) + [('c', 3)] }}", dict(d1=dict(a=1)), [['a', 1], ['c', 3]]),
    ("{{ [('c', 3)] + (d1 | items) }}", dict(d1=dict(a=1)), [['c', 3], ['a', 1]]),
    ("{{ (d1 | items) * 2 }}", dict(d1=dict(a=1)), [['a', 1], ['a', 1]]),
    ("{{ 2 * (d1 | items) }}", dict(d1=dict(a=1)), [['a', 1], ['a', 1]]),
    ("{{ (1, 2) }}", dict(), [1, 2]),
])
def test_lazy_list_adapter_operators(template, variables, expected) -> None:
    templar = TemplateEngine(variables=variables)

    assert templar.template(TRUST.tag(template)) == expected


@pytest.mark.parametrize("expression, expected_value, expected_type", [
    ("l1 + l2", [_LazyValue(1), _LazyValue(2)], _AnsibleLazyTemplateList),  # __add__
    ("l1 + l2f()", [1, 2], list),  # __add__ (different lazy options)
    ("l1 + [2]", [_LazyValue(1), 2], _AnsibleLazyTemplateList),  # __add__
    ("[1] + l2", [1, _LazyValue(2)], _AnsibleLazyTemplateList),  # __radd__
    ("l1 * 3", [_LazyValue(1), _LazyValue(1), _LazyValue(1)], _AnsibleLazyTemplateList),  # __mul__
    ("3 * l2", [_LazyValue(2), _LazyValue(2), _LazyValue(2)], _AnsibleLazyTemplateList),  # __rmul__
    ("d1 | d2", dict(a=1, b=2, c=2), dict),  # __or__
    ("d1 | {'b': 2, 'c': 2}", dict(a=1, b=2, c=2), dict),  # __or__
    ("{'a': 1} | d2", dict(a=1, b=2, c=2), dict),  # __ror__
    ("d1 |= d2", dict(d1=dict(a=_LazyValue(1), b=2, c=2)), _AnsibleLazyTemplateDict),  # __ior__
    ("d1 |= {'b': 2, 'c': 2}", dict(d1=dict(a=_LazyValue(1), b=2, c=2)), _AnsibleLazyTemplateDict),  # __ior__
    ('d1.copy()', dict(a=_LazyValue(1), c=_LazyValue(1)), _AnsibleLazyTemplateDict),  # _AnsibleLazyTemplateDict.copy
    ('type(d1)(d1)', dict(a=_LazyValue(1), c=_LazyValue(1)), _AnsibleLazyTemplateDict),  # _AnsibleLazyTemplateDict.__init__ copy
    ('l1.copy()', [_LazyValue(1)], _AnsibleLazyTemplateList),  # _AnsibleLazyTemplateList.copy
    ('type(l1)(l1)', [_LazyValue(1)], _AnsibleLazyTemplateList),  # _AnsibleLazyTemplateList.__init__ copy
    ('type(t1)(t1)', (1,), _AnsibleLazyAccessTuple),
    ('copy.copy(l1)', [_LazyValue(1)], _AnsibleLazyTemplateList),
    ('copy.copy(d1)', dict(a=_LazyValue(1), c=_LazyValue(1)), _AnsibleLazyTemplateDict),
    ('copy.deepcopy(l1)', [_LazyValue(1)], _AnsibleLazyTemplateList),  # __AnsibleLazyTemplateList.__deepcopy__
    ('copy.deepcopy(d1)', dict(a=_LazyValue(1), c=_LazyValue(1)), _AnsibleLazyTemplateDict),  # __AnsibleLazyTemplateDict.__deepcopy__
    ('ExampleSingletonTag().tag(l1)', [_LazyValue(1)], _AnsibleLazyTemplateList),
    ('ExampleSingletonTag().tag(d1)', dict(a=_LazyValue(1), c=_LazyValue(1)), _AnsibleLazyTemplateDict),
    ('list(reversed(l1))', [1], list),  # _AnsibleLazyTemplateList.__reversed__
    ('list(reversed(d1))', ['c', 'a'], list),  # dict.__reversed__ - keys only
    ('l1[:]', [_LazyValue(1)], _AnsibleLazyTemplateList),  # __getitem__ (slice)
    ('t1[:]', (1,), _AnsibleLazyAccessTuple),  # __getitem__ (slice)
    ('d1["a"]', 1, int),  # __getitem__
    ('d1.get("a")', 1, int),  # get
    ('l1[0]', 1, int),  # __getitem__
    ('d1.pop("a")', 1, int),  # dict.pop (check returned value)
    ('d1.pop("a");', dict(d1=dict(c=_LazyValue(1))), _AnsibleLazyTemplateDict),  # dict.pop (check mutated source)
    ('list(d1.items())', [('a', 1), ('c', 1)], list),  # items
    ('d1.popitem()', ('c', 1), tuple),  # dict.popitem (check returned value)
    ('d1.popitem();', dict(d1=dict(a=_LazyValue(1))), _AnsibleLazyTemplateDict),  # dict.popitem (check mutated source)
    ('d1.clear(); d1["d"] = 4', dict(d1=dict(d=4)), _AnsibleLazyTemplateDict),  # dict.clear (clear + mutate)
    ('d1["d"] = 4; d1.clear()', dict(d1=dict()), _AnsibleLazyTemplateDict),  # dict.clear (mutate + clear)
    ('d1.update(d2);', dict(d1=dict(a=_LazyValue(1), b=2, c=2)), _AnsibleLazyTemplateDict),  # dict.update
    ('d1.setdefault("d", 4)', 4, int),  # dict.setdefault (return value check only)
    ('d1.setdefault("d", 4);', dict(d1=dict(a=_LazyValue(1), c=_LazyValue(1), d=4)), _AnsibleLazyTemplateDict),  # dict.setdefault
    ('l1.clear(); l1.append(4)', dict(l1=[4]), _AnsibleLazyTemplateList),  # list.clear (clear + mutate)
    ('l1.append(4); l1.clear()', dict(l1=[]), _AnsibleLazyTemplateList),  # list.clear (mutate + clear)
    ('l1.pop()', 1, int),  # list.pop (check returned value)
    ('l1.pop();', dict(l1=[]), _AnsibleLazyTemplateList),  # list.pop (check returned value)
    ('l1.insert(0, 4);', dict(l1=[4, _LazyValue(1)]), _AnsibleLazyTemplateList),  # list.insert (check mutated source)
    ('l1.extend(l2);', dict(l1=[_LazyValue(1), 2]), _AnsibleLazyTemplateList),  # list.extend (check mutated source)
    ('l1[0] = 2;', dict(l1=[2]), _AnsibleLazyTemplateList),  # list.__setitem__ (check mutated source)
    ('l1 > [1]', False, bool),  # __gt__ (lazy vs constant)
    ('l1 > l1x', False, bool),  # __gt__ (lazy vs lazy)
    ('l1 >= [1]', True, bool),  # __ge__ (lazy vs constant)
    ('l1 >= l1x', True, bool),  # __ge__ (lazy vs lazy)
    ('l1 < [1]', False, bool),  # __lt__ (lazy vs constant)
    ('l1 < l1x', False, bool),  # __lt__ (lazy vs lazy)
    ('l1 <= [1]', True, bool),  # __le__ (lazy vs constant)
    ('l1 <= l1x', True, bool),  # __le__ (lazy vs lazy)
    ('l1 == [1]', True, bool),  # __eq__ (lazy vs constant)
    ('l1 == l1x', True, bool),  # __eq__ (lazy vs lazy)
    ('l1 != [1]', False, bool),  # __ne__ (lazy vs constant)
    ('l1 != l1x', False, bool),  # __ne__ (lazy vs lazy)
    ('d1 == {"a": 1, "c": 1}', True, bool),  # __eq__ (lazy vs constant)
    ('d1 == d1x', True, bool),  # __eq__ (lazy vs lazy)
    ('d1 != {"a": 1, "c": 1}', False, bool),  # __ne__ (lazy vs constant)
    ('d1 != d1x', False, bool),  # __eq__ (lazy vs lazy)
    ('1 in l1', True, bool),  # __contains__
    ('str(d1)', str(dict(a=1, c=1)), str),  # __str__
    ('str(l1)', str([1]), str),  # __str__
    ('repr(d1)', repr(dict(a=1, c=1)), str),  # __repr__
    ('repr(l1)', repr([1]), str),  # __repr__
    ('"{}".format(l1)', repr([1]), str),  # __format__ [no override required]
    ('list(l1)', [1], list),  # _AnsibleLazyTemplateList.__iter__
    ('list(d1)', ['a', 'c'], list),  # _AnsibleLazyTemplateDict.__iter__
    ('l1._native_copy()', [1], list),  # native_copy
    ('dict(d1)', dict(a=1, c=1), dict),  # __iter__
    ('d1._native_copy()', dict(a=1, c=1), dict),  # native_copy
    ('{} + l1', "unsupported operand type(s) for +: 'dict' and '_AnsibleLazyTemplateList'", TypeError),  # __radd__
    ('d1 + l1', "unsupported operand type(s) for +: '_AnsibleLazyTemplateDict' and '_AnsibleLazyTemplateList'", TypeError),  # __radd__
    ('d1 + []', "unsupported operand type(s) for +: '_AnsibleLazyTemplateDict' and 'list'", TypeError),  # python operator dispatch
    ('set() + l1', "unsupported operand type(s) for +: 'set' and '_AnsibleLazyTemplateList'", TypeError),  # python operator dispatch
    ('set() + d1', "unsupported operand type(s) for +: 'set' and '_AnsibleLazyTemplateDict'", TypeError),  # python operator dispatch
    ('l1 + {}', 'can only concatenate list (not "dict") to list', TypeError),  # __add__ (relies on list.__add__)
    ('l1 + set()', 'can only concatenate list (not "set") to list', TypeError),  # __add__ (relies on list.__add__)
    ('l1 + tuple()', 'can only concatenate list (not "tuple") to list', TypeError),  # __add__ (relies on list.__add__)
    ('tuple() + l1', 'can only concatenate tuple (not "_AnsibleLazyTemplateList") to tuple', TypeError),  # __radd__ (relies on tuple.__add__)
    ('tuple() + d1', 'can only concatenate tuple (not "_AnsibleLazyTemplateDict") to tuple', TypeError),  # relies on tuple.__add__
    ('l1.pop(42)', "pop index out of range", IndexError),
    ('type(l1)([])', 'Direct construction of lazy containers is not supported.', UnsupportedConstructionMethodError),
    ('type(t1)([])', 'Direct construction of lazy containers is not supported.', UnsupportedConstructionMethodError),
    ('type(d1)({})', 'Direct construction of lazy containers is not supported.', UnsupportedConstructionMethodError),
], ids=str)
def test_lazy_container_operators(expression: str, expected_value: t.Any, expected_type: type) -> None:
    """
    Verify that lazy container operators and methods properly implement lazy behavior.
    Results are checked both for expected types and expected values.
    When the result is a container, items in the container are checked to see if they're lazy as appropriate.
    This test uses a function to simulate Jinja plugin behavior, since plugins can use operators and methods that Jinja expressions cannot.
    """
    # DTFIX5: add a unit test to ensure every list/dict method has been overridden or on a list we can safely ignore
    def l2f() -> list:
        """Return a lazy list that uses different lazy options, to ensure it cannot be lazy combined."""
        return TemplateContext.current().templar.template([2], lazy_options=LazyOptions.SKIP_TEMPLATES)

    variables = dict(
        one=1,
        two=2,
        data=dict(
            l1=[TRUST.tag('{{ one }}')],
            l1x=[TRUST.tag('{{ one }}')],
            l2=[TRUST.tag('{{ two }}')],
            l2f=l2f,
            t1=(TRUST.tag('{{ one }}'),),
            d1=dict(a=TRUST.tag('{{ one }}'), c=TRUST.tag('{{ one }}')),
            d1x=dict(a=TRUST.tag('{{ one }}'), c=TRUST.tag('{{ one }}')),
            d2=dict(b=TRUST.tag('{{ two }}'), c=TRUST.tag('{{ two }}')),
        ),
    )

    def run_test(data: dict[str, t.Any]) -> t.Any:
        nonlocal expected_value

        secondary_templar = TemplateEngine()

        # Run under a secondary context using a templar with no variables; this allows us to test the correct propagation and use of the
        # embedded templar in lazy containers. Templated values will not render correctly if they pick up the ambient (no-vars) templar during
        # various copy/operator scenarios.
        with TemplateContext(template_value='', templar=secondary_templar, options=TemplateOptions.DEFAULT, stop_on_template=False):
            code_globals = dict(
                copy=copy,
                ExampleSingletonTag=ExampleSingletonTag,
            )

            try:
                result = eval(expression, code_globals, data)
            except SyntaxError:
                # some expressions use a semicolon to force exec instead of eval, even if they only need a single statement
                exec(expression, code_globals, data)

                var_name = list(expected_value)[0]
                expected_value = expected_value[var_name]
                result = data[var_name]
            except Exception as ex:
                if type(ex) is not expected_type:  # pylint: disable=unidiomatic-typecheck
                    # we weren't expecting an exception, or got one of the wrong type; re-raise it now for the traceback instead of just a failed assertion
                    raise

                result = ex

        assert type(result) is expected_type  # pylint: disable=unidiomatic-typecheck

        expected_result: t.Any  # avoid type narrowing

        if issubclass(expected_type, list):
            assert isinstance(result, list)  # redundant, but assists mypy in understanding the type

            expected_list_types = [type(value) for value in expected_value]
            expected_result = [value.value if isinstance(value, _LazyValue) else value for value in expected_value]

            actual_list_types: list[type] = [type(value) for value in list.__iter__(result)]

            assert actual_list_types == expected_list_types
        elif issubclass(expected_type, tuple):
            assert isinstance(result, tuple)  # redundant, but assists mypy in understanding the type

            expected_tuple_types = [type(value) for value in expected_value]
            expected_result = expected_value

            actual_tuple_types: list[type] = [type(value) for value in tuple.__iter__(result)]

            assert actual_tuple_types == expected_tuple_types
        elif issubclass(expected_type, dict):
            assert isinstance(result, dict)  # redundant, but assists mypy in understanding the type

            expected_dict_types = {key: type(value) for key, value in expected_value.items()}
            expected_result = {key: value.value if isinstance(value, _LazyValue) else value for key, value in expected_value.items()}

            actual_dict_types: dict[str, type] = {key: type(value) for key, value in dict.items(result)}

            assert actual_dict_types == expected_dict_types
        elif issubclass(expected_type, Exception):
            result = str(result)  # unfortunately exceptions can't be compared for equality, so use the string representation instead
            expected_result = expected_value
        else:
            expected_result = expected_value

        assert result == expected_result

    templar = TemplateEngine(variables=variables)
    templar.environment.globals['run_test'] = run_test
    templar.template(TRUST.tag('{{ run_test(data=data) }}'))


def test_range_templating():
    """
    Verify special handling for Range objects.
    They are usually listified like other iterables when returned from a Jinja filter or method/function call, except when calling the Python `range()`
    global function directly, which allows the range object to be returned and used directly.
    """
    templar = TemplateEngine(variables=dict(
        large_value=min(1000000000000, sys.maxsize - 1)  # ensure we don't exceed ssize_t on 32-bit systems
    ))

    # ensure that an insanely large range is not listified
    assert templar.evaluate_conditional(TRUST.tag("range(large_value) | type_debug == 'range'"))
    assert isinstance(templar.template(TRUST.tag("{{ range(large_value) | random }}")), int)
    assert templar.template(TRUST.tag("{{ range(3) | reverse }}")) == [2, 1, 0]


@pytest.mark.parametrize("value", [
    "{{ range(10000) }}",
    "{{ [range(10000)] }}",
    "{{ {'a': range(10000)} }}",
])
def test_range_template_fail(value):
    """Verify unsupported range object usages."""
    with pytest.raises(AnsibleTemplateError):
        TemplateEngine().template(TRUST.tag(value))


LISTIFIED_ITERATOR_VALUES_AND_EXPECTED = (
    ("{'a': 1, 'b': 2}.items()", [('a', 1), ('b', 2)]),
    ("['hi'] | map('upper')", ["HI"]),
)

LISTIFIED_ITERATOR_VALUES = tuple(item[0] for item in LISTIFIED_ITERATOR_VALUES_AND_EXPECTED)
LISTIFIED_ITERATOR_TEMPLATES = tuple(TRUST.tag(f'{{{{ {value} }}}}') for value in LISTIFIED_ITERATOR_VALUES)


@pytest.mark.parametrize("value, expected", LISTIFIED_ITERATOR_VALUES_AND_EXPECTED)
def test_list_adapter_equality(value: str, expected: list) -> None:
    origin = Origin(path='/test')  # here to make sure it doesn't trigger an exception, it won't be in the result

    assert TemplateEngine().evaluate_expression(TRUST.tag(origin.tag(f"{value} == {expected}")))


@pytest.mark.parametrize("value", LISTIFIED_ITERATOR_VALUES)
def test_list_adapter_source_propagation(value: t.Any) -> None:
    origin = Origin(path='/test')

    assert Origin.get_tag(TemplateEngine().template(TRUST.tag(origin.tag(f"{{{{ {value} }}}}")))) is origin


@pytest.mark.parametrize("value", t.cast(tuple, CONTAINER_VALUES) + LISTIFIED_ITERATOR_TEMPLATES, ids=str)
def test_lazy_containers_to_yaml(value: t.Any) -> None:
    TemplateEngine(variables=dict(value=value)).template(TRUST.tag("{{ value | to_yaml }}"))


@pytest.mark.parametrize("value", t.cast(tuple, CONTAINER_VALUES) + LISTIFIED_ITERATOR_TEMPLATES, ids=str)
def test_lazy_containers_to_json(value: t.Any) -> None:
    TemplateEngine(variables=dict(value=value)).template(TRUST.tag("{{ value | to_json }}"))


@pytest.mark.parametrize("value", (_AnsibleLazyTemplateDict, _AnsibleLazyTemplateList))
def test_lazy_interface(value: type[_AnsibleLazyTemplateMixin]) -> None:
    """Ensure that lazy containers implement all the methods that their native types provide."""
    missing = set(dir(value._native_type)) - set(dir(value))
    assert not missing


@pytest.mark.parametrize("expression, expected", (
    ("l1[0]", [0]),  # _AnsibleLazyTemplateList.__getitem__ and _ansible_finalize (list to lazy list)
    ("l1[1]", [1]),  # _AnsibleLazyTemplateList.__getitem__
    ("l1[2]", 'hello'),  # _AnsibleLazyTemplateList.__getitem__
    ("[v for v in l1][0]", [0]),  # _AnsibleLazyTemplateList.__iter__ and _ansible_finalize (list to lazy list)
    ("[v for v in l1][1]", [1]),  # _AnsibleLazyTemplateList.__iter__
    ("[v for v in l1][2]", 'hello'),  # _AnsibleLazyTemplateList.__iter__
    ("d1['a']", [0]),  # _AnsibleLazyTemplateDict.__getitem__ and _ansible_finalize (dict to lazy dict)
    ("d1['b']", [1]),  # _AnsibleLazyTemplateDict.__getitem__
    ("d1['c']", 'hello'),  # _AnsibleLazyTemplateDict.__getitem__
    ("{k: v for k, v in d1.items()}['a']", [0]),  # _AnsibleLazyTemplateDict.items and _ansible_finalize (dict to lazy dict)
    ("{k: v for k, v in d1.items()}['b']", [1]),  # _AnsibleLazyTemplateDict.items
    ("{k: v for k, v in d1.items()}['c']", 'hello'),  # _AnsibleLazyTemplateDict.items
), ids=str)
def test_lazy_persistence(expression: str, expected: t.Any) -> None:
    """
    Verify that values returned from lazy containers are persistent, regardless of whether templating is involved or not.
    A global function is used to simulate the behavior that would occur in a Jinja plugin.
    """
    variables = dict(
        data=dict(
            l1=[TRUST.tag('{{ [0] }}'), [1], 'hello'],
            d1=dict(a=TRUST.tag('{{ [0] }}'), b=[1], c='hello'),
        ),
    )

    def run_test(data: dict[str, t.Any]) -> None:
        first = eval(expression, data)
        second = eval(expression, data)

        assert first == expected
        assert first is second

    templar = TemplateEngine(variables=variables)
    templar.environment.globals['run_test'] = run_test

    templar.template(TRUST.tag("{{ run_test(data=data) }}"))


@pytest.mark.parametrize("expression", (
    "data['l1'][0]",
    "data['d1']['a']",
))
def test_lazy_mutation_persistence(expression: str) -> None:
    """Verify that lazy containers persist values added after creation and return them as-is without modification, even if they contain trusted templates."""
    # DTFIX5: investigate relevance of this test now that mutation/dirty tracking is toast
    variables = dict(
        data=dict(
            l1=[None],
            d1=dict(a=None),
        ),
    )

    def run_test(data: dict[str, t.Any]) -> None:
        assert data

        value = [TRUST.tag('{{ "hi" }}')]

        exec(f'{expression} = value')

        result = eval(expression)

        assert result is value

    templar = TemplateEngine(variables=variables)
    templar.environment.globals['run_test'] = run_test

    templar.template(TRUST.tag("{{ run_test(data=data) }}"))


@pytest.mark.parametrize("expr, new_value, some_var, expected_value", [
    ("access_and_mutate_dict(mutate_dict(some_var))", TRUST.tag("{{ 'mom' }}"), dict(one="one"), dict(one="one", new="{{ 'mom' }}", secondnew="{{ 'mom' }}")),
    ("access_and_mutate_list(mutate_list(some_var))", TRUST.tag("{{ 'mom' }}"), ["one"], ["one", "{{ 'mom' }}", "{{ 'mom' }}"]),
])
def test_lazy_mutation_cross_plugin_dirty_container(expr: str, new_value: t.Any, some_var: t.Any, expected_value: t.Any):
    """Ensure that new templates sourced from a plugin are not processed by subsequent plugins or template finalization."""
    # DTFIX5: investigate relevance of this test now that mutation/dirty tracking is toast
    def mutate_list(value: list) -> list:
        value.append(new_value)

        assert value[1] is new_value

        return value

    def access_and_mutate_list(value: list) -> list:
        value.append(new_value)

        assert value[2] is new_value
        assert value[1] == expected_value[1]

        return value

    def mutate_dict(value: dict) -> dict:
        value['new'] = new_value

        assert value['new'] is new_value

        return value

    def access_and_mutate_dict(value: t.Any) -> list:
        value['secondnew'] = new_value

        assert value['secondnew'] is new_value
        assert value['new'] == expected_value['new']

        return value

    available_vars = dict(some_var=some_var,
                          access_and_mutate_list=access_and_mutate_list,
                          mutate_list=mutate_list,
                          access_and_mutate_dict=access_and_mutate_dict,
                          mutate_dict=mutate_dict)

    res = TemplateEngine(variables=available_vars).evaluate_expression(TRUST.tag(expr))
    assert res == expected_value


def test_undefined_in_jinja_constant_container():
    """
    Retrieval of an undefined value in a plugin which does not declare undefined support should internally raise AnsibleUndefinedVariable, which, if
    unhandled by the plugin, will be converted to an AnsibleUndefined and returned as the result of that plugin's invocation, allowing the template
    to continue execution.
    """

    plugin_retrieved_the_value = False

    def demo(_input, arg):
        nonlocal plugin_retrieved_the_value
        _x = arg[0]
        plugin_retrieved_the_value = True
        return []

    templar = TemplateEngine()
    templar.environment.filters['demo'] = demo

    with pytest.raises(AnsibleUndefinedVariable):
        x = templar.template(TRUST.tag("{{ True | demo([bogus_var]) }}"))
        pass
    assert not plugin_retrieved_the_value

    assert templar.template(TRUST.tag("{{ True | demo([bogus_var]) | default('nope') }}")) == 'nope'
    assert not plugin_retrieved_the_value


class CustomSequence(c.Sequence):
    def __init__(self, values) -> None:
        self._content = list(values)

    def __getitem__(self, item) -> t.Any:
        return self._content[item]

    def __len__(self) -> int:
        return len(self._content)


@pytest.mark.parametrize("expression, expected_result", (
    # verify templates sourced by plugins are not auto-templated
    ("list_with_bad_template() | pass_through is first_item_unrendered", True),  # plain list return values are lazy wrapped, but not templated
    ("tuple_with_bad_template() | pass_through is first_item_unrendered", True),  # tuples are not lazified, but suppressed unsupported type warning
    # verify markers are tripped on managed access, but not unmanaged access (they will still trip on use, which is not tested here)
    ("list_with_undefined() | pass_through is first_item_trips", True),  # managed access trips
    ("tuple_with_undefined() | pass_through is first_item_trips", True),  # managed access trips
    ("custom_sequence_with_undefined() | pass_through is first_item_trips", False),  # unmanaged access doesn't trip
    # verify that Jinja constant containers are wrapped where possible
    ("['{{ 1 }}'] | pass_through is first_item_unrendered", True),
    ("[bogusvar] | pass_through is first_item_trips", True),
    ("{'k': bogusvar} | pass_through is first_item_trips", True),
    ("(bogusvar,) | pass_through is first_item_trips", True),
    # verify that plugins directly invoking tests and filters do not trigger auto-templating
    ('call_filter_with_native_args_kwargs()', True),  # `call_filter` invocations with plain list/dict should be lazy non-templating
    ('call_test_with_native_args_kwargs()', True),  # `call_test` invocations with plain list/dict should be lazy non-templating
))
def test_plugin_result_wrapping(expression: str, expected_result: t.Any, _ignore_untrusted_template) -> None:
    """
    Validate various intra-plugin container behaviors:
     * A plain list/dict returned by a Jinja call/plugin gets lazified, but with templating disabled (feature parity with pre-2.19).
     * Tuple/set returned by a Jinja call/plugin receives no special behavior.
     * Jinja plugins that do not accept markers should raise when accessing one from a lazy container.
    The default test "untrusted template as error" behavior is disabled, since some test cases involve accessing untrusted templates.
    """
    templar = TemplateEngine()

    def list_with_bad_template() -> list[str]:
        return [TRUST.tag('{{ 1 / 0 }}')]

    def list_with_undefined() -> list[Marker]:
        return [TemplateEngine().template(TRUST.tag('{{ bogusvar }}'))]

    def tuple_with_bad_template() -> tuple[str, ...]:
        return (TRUST.tag('{{ 1 / 0 }}'),)

    def tuple_with_undefined() -> tuple[Marker, ...]:
        return (TemplateEngine().template(TRUST.tag('{{ bogusvar }}')),)

    def custom_sequence_with_undefined() -> CustomSequence:
        return CustomSequence((TemplateEngine().template(TRUST.tag('{{ bogusvar }}')),))

    def pass_through(value: t.Any) -> t.Any:
        return value

    def first_item_trips(value: c.Sequence | c.Mapping) -> bool:
        if isinstance(value, c.Mapping):
            key_or_idx = list(value.keys())[0]
        else:
            key_or_idx = 0

        try:
            value[key_or_idx]
        except MarkerError:
            return True

        return False

    def first_item_unrendered(value: c.Sequence) -> bool:
        return isinstance(value[0], str) and is_possibly_template(value[0])

    def call_filter_with_native_args_kwargs() -> t.Any:
        return templar.environment.call_filter('args_and_kwargs_are_non_templating_lazy', value=42, args=[[1]], kwargs=dict(kwarg=[2]))

    def call_test_with_native_args_kwargs() -> t.Any:
        return templar.environment.call_test('args_and_kwargs_are_non_templating_lazy', value=42, args=[[1]], kwargs=dict(kwarg=[2]))

    def args_and_kwargs_are_non_templating_lazy(value: t.Any, arg, *, kwarg) -> bool:
        return (
            value == 42 and
            isinstance(arg, _AnsibleLazyTemplateList) and
            not arg._lazy_options.template and
            isinstance(kwarg, _AnsibleLazyTemplateList) and
            not kwarg._lazy_options.template
        )

    templar.environment.globals.update(
        list_with_bad_template=list_with_bad_template,
        list_with_undefined=list_with_undefined,
        tuple_with_bad_template=tuple_with_bad_template,
        tuple_with_undefined=tuple_with_undefined,
        custom_sequence_with_undefined=custom_sequence_with_undefined,
        call_filter_with_native_args_kwargs=call_filter_with_native_args_kwargs,
        call_test_with_native_args_kwargs=call_test_with_native_args_kwargs,
    )

    templar.environment.filters.update(
        pass_through=pass_through,
        args_and_kwargs_are_non_templating_lazy=args_and_kwargs_are_non_templating_lazy,
    )

    templar.environment.tests.update(
        first_item_unrendered=first_item_unrendered,
        first_item_trips=first_item_trips,
        args_and_kwargs_are_non_templating_lazy=args_and_kwargs_are_non_templating_lazy,
    )

    expression = TRUST.tag(expression)

    with emits_warnings(warning_pattern=[], deprecation_pattern=[]):
        result = templar.evaluate_expression(expression)

    assert result == expected_result


def test_lazy_options_propagation(template_context):
    """Ensure that lazy collections propagate lazy options to child lazies."""
    src_template = "{{ 1 / 0 }}"

    def nested_container_with_template() -> t.Any:
        return [[TRUST.tag(src_template)]]

    templar = TemplateEngine()
    templar.environment.globals.update(nested_container_with_template=nested_container_with_template)

    res = templar.template(TRUST.tag('{{ nested_container_with_template() }}'))

    assert isinstance(res, _AnsibleLazyTemplateList)
    assert not res._lazy_options.template  # ensure call output skips templating
    assert isinstance(res[0], _AnsibleLazyTemplateList)
    assert res[0]._lazy_options is res._lazy_options  # ensure the parent options are propagated
    assert res[0][0] == src_template


def test_subclass_dispatch_collision():
    """Simulate a developer error by creating a new lazy container subclass with colliding base and/or tagged type dispatch entries."""

    class _TestBaseType: ...
    class _TaggedTestBaseType(_TestBaseType, AnsibleTaggedObject): ...
    class _LazyTaggedTestBaseType(_TaggedTestBaseType, _AnsibleLazyTemplateMixin): ...

    with pytest.raises(
        TypeError,
        match=f"Lazy mixin '_OopsConflictingLazyTaggedTestBaseType' type '_TaggedTestBaseType' conflicts with {_LazyTaggedTestBaseType.__name__!r}.",
    ):
        class _OopsConflictingLazyTaggedTestBaseType(_TaggedTestBaseType, _AnsibleLazyTemplateMixin): ...
        assert _OopsConflictingLazyTaggedTestBaseType  # pragma: nocover

    with pytest.raises(
        TypeError,
        match=f"Lazy mixin '_OopsConflictingTestBaseType' type '_TestBaseType' conflicts with {_LazyTaggedTestBaseType.__name__!r}.",
    ):
        class _OopsConflictingTestBaseType(_TestBaseType, _AnsibleLazyTemplateMixin): ...
        assert _OopsConflictingTestBaseType  # pragma: nocover


_L = ['a']
_D = dict(B='b')
_T = ('c',)


@pytest.mark.parametrize("value, deep", (
    ([_L, _D, _T], False),
    ([_L, _D, _T], True),
    (dict(L=_L, D=_D, T=_T), False),
    (dict(L=_L, D=_D, T=_T), True),
    ((_L, _D, _T), False),
    ((_L, _D, _T), True),
), ids=str)
def test_lazy_copies(value: list | dict, deep: bool, template_context: TemplateContext) -> None:
    """Verify that `copy.copy` and `copy.deepcopy` make lazy copies of lazy containers."""
    # pylint: disable=unnecessary-dunder-call
    original = _AnsibleLazyTemplateMixin._try_create(value)
    base_type = type(value)
    pseudo_lazy = base_type is tuple  # tuples do not wrap their values in `_LazyValue`

    keys: list

    if base_type is list or base_type is tuple:
        keys = list(range(len(original)))
    else:
        keys = list(original)

    assert all(isinstance(base_type.__getitem__(original, key), _LazyValue) != pseudo_lazy for key in keys)  # lazy before copy

    if deep:
        copied = copy.deepcopy(original)
    else:
        copied = copy.copy(original)

    assert copied is not original
    assert len(copied) == len(original)
    assert pseudo_lazy or all(isinstance(base_type.__getitem__(original, key), _LazyValue) != pseudo_lazy for key in keys)  # still lazy after copy

    assert all((base_type.__getitem__(copied, key) is base_type.__getitem__(original, key)) != deep for key in keys)
    assert (copied._templar is original._templar) != deep
    assert (copied._lazy_options is original._lazy_options) != deep


def test_lazy_template_mixin_init() -> None:
    """
    Verify `_AnsibleLazyTemplateMixin` checks the __init__ arg type.
    This code path is not normally reachable, since types which use it perform the same check before invoking the mixin.
    """
    with pytest.raises(UnsupportedConstructionMethodError):
        _AnsibleLazyTemplateMixin(t.cast(t.Any, None))



# Copyright (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = '''
    name: script
    version_added: "2.4"
    short_description: Executes an inventory script that returns JSON
    options:
      always_show_stderr:
        description: Toggle display of stderr even when script was successful
        version_added: "2.5.1"
        default: True
        type: boolean
        ini:
           - section: inventory_plugin_script
             key: always_show_stderr
        env:
           - name: ANSIBLE_INVENTORY_PLUGIN_SCRIPT_STDERR
    description:
        - The source provided must be an executable that returns Ansible inventory JSON
        - The source must accept C(--list) and C(--host <hostname>) as arguments.
          C(--host) will only be used if no C(_meta) key is present.
          This is a performance optimization as the script would be called one additional time per host otherwise.
    notes:
        - Enabled in configuration by default.
        - The plugin does not cache results because external inventory scripts are responsible for their own caching.
        - To write your own inventory script see (R(Developing dynamic inventory,developing_inventory) from the documentation site.
        - To find the scripts that used to be part of the code release, go to U(https://github.com/ansible-community/contrib-scripts/).
        - Since 2.19 using a directory as an inventory source will no longer ignore .ini files by default,
          but you can still update the configuration to do so.
'''

EXAMPLES = r'''# fmt: code

### simple bash script

   #!/usr/bin/env bash

   if [ "$1" == "--list" ]; then
   cat<<EOF
   {
     "bash_hosts": {
       "hosts": [
         "myhost.domain.com",
         "myhost2.domain.com"
       ],
       "vars": {
         "host_test": "test-value"
       }
     },
     "_meta": {
       "hostvars": {
         "myhost.domain.com": {
           "host_specific_test_var": "test-value"
         }
       }
     }
   }
   EOF
   elif [ "$1" == "--host" ]; then
     # this should not normally be called by Ansible as we return _meta above
     if [ "$2" == "myhost.domain.com" ]; then
        echo '{"_meta": {hostvars": {"myhost.domain.com": {"host_specific-test_var": "test-value"}}}}'
     else
        echo '{"_meta": {hostvars": {}}}'
     fi
   else
     echo "Invalid option: use --list or --host <hostname>"
     exit 1
   fi


### python example with ini config

    #!/usr/bin/env python
    """
    # ansible_inventory.py
    """
    import argparse
    import json
    import os.path
    import sys
    from configparser import ConfigParser
    from inventories.custom import MyInventoryAPI

    def load_config() -> ConfigParser:
        cp = ConfigParser()
        config_file = os.path.expanduser("~/.config/ansible_inventory_script.cfg")
        cp.read(config_file)
        if not cp.has_option('DEFAULT', 'namespace'):
            raise ValueError("Missing configuration option: DEFAULT -> namespace")
        return cp


    def get_api_data(namespace: str, pretty=False) -> str:
        """
        :param namespace: parameter for our custom api
        :param pretty: Human readable JSON vs machine readable
        :return: JSON string
        """
        found_data = list(MyInventoryAPI(namespace))
        hostvars = {}
        data = { '_meta': { 'hostvars': {}},}

        groups = found_data['groups'].keys()
        for group in groups:
            groups[group]['hosts'] = found_data[groups].get('host_list', [])
            if group not in data:
                data[group] = {}
            data[group]['hosts'] = found_data[groups].get('host_list', [])
            data[group]['vars'] = found_data[groups].get('info', [])
            data[group]['children'] = found_data[group].get('subgroups', [])

        for host_data in found_data['hosts']:
            for name in host_data.items():
                # turn info into vars
                data['_meta'][name] = found_data[name].get('info', {})
                # set ansible_host if possible
                if 'address' in found_data[name]:
                    data[name]['_meta']['ansible_host'] = found_data[name]['address']
        data['_meta']['hostvars'] = hostvars

        return json.dumps(data, indent=pretty)

    if __name__ == '__main__':

        arg_parser = argparse.ArgumentParser( description=__doc__, prog=__file__)
        arg_parser.add_argument('--pretty', action='store_true', default=False, help="Pretty JSON")
        mandatory_options = arg_parser.add_mutually_exclusive_group()
        mandatory_options.add_argument('--list', action='store', nargs="*", help="Get inventory JSON from our API")
        mandatory_options.add_argument('--host', action='store',
                                       help="Get variables for specific host, not used but kept for compatibility")

        try:
            config = load_config()
            namespace = config.get('DEFAULT', 'namespace')

            args = arg_parser.parse_args()
            if args.host:
                print('{"_meta":{}}')
                sys.stderr.write('This script already provides _meta via --list, so this option is really ignored')
            elif len(args.list) >= 0:
                print(get_api_data(namespace, args.pretty))
            else:
                raise ValueError("Valid options are --list or --host <HOSTNAME>")

        except ValueError:
            raise

'''


import json
import os
import shlex
import subprocess
import typing as t

from ansible.errors import AnsibleError, AnsibleJSONParserError
from ansible.inventory.data import InventoryData
from ansible.module_utils.datatag import native_type_name
from ansible.module_utils.common.json import get_decoder
from ansible.parsing.dataloader import DataLoader
from ansible.plugins.inventory import BaseInventoryPlugin
from ansible._internal._datatag._tags import TrustedAsTemplate, Origin
from ansible.utils.display import Display
from ansible._internal._json._profiles import _legacy, _inventory_legacy

display = Display()


class InventoryModule(BaseInventoryPlugin):
    """Host inventory parser for ansible using external inventory scripts."""

    NAME = 'script'

    def __init__(self) -> None:
        super(InventoryModule, self).__init__()

        self._hosts: set[str] = set()

    def verify_file(self, path: str) -> bool:
        return super(InventoryModule, self).verify_file(path) and os.access(path, os.X_OK)

    def parse(self, inventory: InventoryData, loader: DataLoader, path: str, cache: bool = False) -> None:
        super(InventoryModule, self).parse(inventory, loader, path)

        self.set_options()

        origin = Origin(description=f'<inventory script output from {path!r}>')

        data, stderr, stderr_help_text = run_command(path, ['--list'], origin)

        try:
            profile_name = detect_profile_name(data)
            decoder = get_decoder(profile_name)
        except Exception as ex:
            raise AnsibleError(
                message="Unable to get JSON decoder for inventory script result.",
                help_text=stderr_help_text,
                # obj will be added by inventory manager
            ) from ex

        try:
            try:
                processed = json.loads(data, cls=decoder)
            except Exception as json_ex:
                AnsibleJSONParserError.handle_exception(json_ex, origin)
        except Exception as ex:
            raise AnsibleError(
                message="Inventory script result could not be parsed as JSON.",
                help_text=stderr_help_text,
                # obj will be added by inventory manager
            ) from ex

        # if no other errors happened, and you want to force displaying stderr, do so now
        if stderr and self.get_option('always_show_stderr'):
            self.display.error(msg=stderr)

        data_from_meta: dict | None = None

        # A "_meta" subelement may contain a variable "hostvars" which contains a hash for each host
        # if this "hostvars" exists at all then do not call --host for each # host.
        # This is for efficiency and scripts should still return data
        # if called with --host for backwards compat with 1.2 and earlier.
        for (group, gdata) in processed.items():
            if group == '_meta':
                data_from_meta = gdata.get('hostvars')

                if not isinstance(data_from_meta, dict):
                    raise TypeError(f"Value contains '_meta.hostvars' which is {native_type_name(data_from_meta)!r} instead of {native_type_name(dict)!r}.")
            else:
                self._parse_group(group, gdata, origin)

        if data_from_meta is None:
            display.deprecated(
                msg="Inventory scripts should always provide 'meta.hostvars'. "
                    "Host variables will be collected by running the inventory script with the '--host' option for each host.",
                version='2.23',
                obj=origin,
            )

        for host in self._hosts:
            if data_from_meta is None:
                got = self.get_host_variables(path, host, origin)
            else:
                got = data_from_meta.get(host, {})

            self._populate_host_vars([host], got)

    def _parse_group(self, group: str, data: t.Any, origin: Origin) -> None:
        """Normalize and ingest host/var information for the named group."""
        group = self.inventory.add_group(group)

        if not isinstance(data, dict):
            original_type = native_type_name(data)
            data = {'hosts': data}
            display.deprecated(
                msg=f"Group {group!r} was converted to {native_type_name(dict)!r} from {original_type!r}.",
                version='2.23',
                obj=origin,
            )
        elif not any(k in data for k in ('hosts', 'vars', 'children')):
            data = {'hosts': [group], 'vars': data}
            display.deprecated(
                msg=f"Treating malformed group {group!r} as the sole host of that group. Variables provided in this manner cannot be templated.",
                version='2.23',
                obj=origin,
            )

        if (data_hosts := data.get('hosts', ...)) is not ...:
            if not isinstance(data_hosts, list):
                raise TypeError(f"Value contains '{group}.hosts' which is {native_type_name(data_hosts)!r} instead of {native_type_name(list)!r}.")

            for hostname in data_hosts:
                self._hosts.add(hostname)
                self.inventory.add_host(hostname, group)

        if (data_vars := data.get('vars', ...)) is not ...:
            if not isinstance(data_vars, dict):
                raise TypeError(f"Value contains '{group}.vars' which is {native_type_name(data_vars)!r} instead of {native_type_name(dict)!r}.")

            for k, v in data_vars.items():
                self.inventory.set_variable(group, k, v)

        if group != '_meta' and isinstance(data, dict) and 'children' in data:
            for child_name in data['children']:
                child_name = self.inventory.add_group(child_name)
                self.inventory.add_child(group, child_name)

    @staticmethod
    def get_host_variables(path: str, host: str, origin: Origin) -> dict:
        """Runs <script> --host <hostname>, to determine additional host variables."""
        origin = origin.replace(description=f'{origin.description} for host {host!r}')

        data, stderr, stderr_help_text = run_command(path, ['--host', host], origin)

        if not data.strip():
            return {}

        try:
            try:
                # Use standard legacy trust inversion here.
                # Unlike the normal inventory output, everything here is considered a variable and thus supports trust (and trust inversion).
                processed = json.loads(data, cls=_legacy.Decoder)
            except Exception as json_ex:
                AnsibleJSONParserError.handle_exception(json_ex, origin)
        except Exception as ex:
            raise AnsibleError(
                message=f"Inventory script result for host {host!r} could not be parsed as JSON.",
                help_text=stderr_help_text,
                # obj will be added by inventory manager
            ) from ex

        return processed


def detect_profile_name(value: str) -> str:
    """
    Detect (optional) JSON profile name from an inventory JSON document.
    Defaults to `inventory_legacy`.
    """
    try:
        data = json.loads(value)
    except Exception as ex:
        raise ValueError('Value could not be parsed as JSON.') from ex

    if not isinstance(data, dict):
        raise TypeError(f'Value is {native_type_name(data)!r} instead of {native_type_name(dict)!r}.')

    if (meta := data.get('_meta', ...)) is ...:
        return _inventory_legacy.Decoder.profile_name

    if not isinstance(meta, dict):
        raise TypeError(f"Value contains '_meta' which is {native_type_name(meta)!r} instead of {native_type_name(dict)!r}.")

    if (profile := meta.get('profile', ...)) is ...:
        return _inventory_legacy.Decoder.profile_name

    if not isinstance(profile, str):
        raise TypeError(f"Value contains '_meta.profile' which is {native_type_name(profile)!r} instead of {native_type_name(str)!r}.")

    if not profile.startswith('inventory_'):
        raise ValueError(f"Non-inventory profile {profile!r} is not allowed.")

    return profile


def run_command(path: str, options: list[str], origin: Origin) -> tuple[str, str, str]:
    """Run an inventory script, normalize and validate output."""
    cmd = [path] + options

    try:
        sp = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as ex:
        raise AnsibleError(
            message=f"Failed to execute inventory script command {shlex.join(cmd)!r}.",
            # obj will be added by inventory manager
        ) from ex

    stdout_bytes, stderr_bytes = sp.communicate()

    stderr = stderr_bytes.decode(errors='surrogateescape')

    if stderr and not stderr.endswith('\n'):
        stderr += '\n'

    # DTFIX-FUTURE: another use case for the "not quite help text, definitely not message" diagnostic output on errors
    stderr_help_text = f'Standard error from inventory script:\n{stderr}' if stderr.strip() else None

    if sp.returncode != 0:
        raise AnsibleError(
            message=f"Inventory script returned non-zero exit code {sp.returncode}.",
            help_text=stderr_help_text,
            # obj will be added by inventory manager
        )

    try:
        data = stdout_bytes.decode()
    except Exception as ex:
        raise AnsibleError(
            "Inventory script result contained characters that cannot be interpreted as UTF-8.",
            help_text=stderr_help_text,
            # obj will be added by inventory manager
        ) from ex
    else:
        data = TrustedAsTemplate().tag(origin.tag(data))

    return data, stderr, stderr_help_text



from __future__ import annotations

import pathlib
import tempfile

from unittest.mock import call

import typing as t

import pytest
import pytest_mock

from ansible import constants as C
from ansible._internal._errors._error_utils import format_exception_message
from ansible._internal._datatag._tags import Origin
from ansible.parsing.utils.yaml import from_yaml
from ansible._internal._yaml._errors import AnsibleYAMLParserError
from ansible.utils.display import Display


@pytest.mark.parametrize("content, expected_message, expect_help_text, line, col", (
    # cases with multiple permutations due to handling of multiple levels of list and up to one dict level

    ('{{ bar }}', 'This may be an issue with missing quotes around a template block.', True, 1, 1),
    ('foo: {{ bar }}', 'This may be an issue with missing quotes around a template block.', True, 1, 6),
    ('- {{ bar }}', 'This may be an issue with missing quotes around a template block.', True, 1, 3),
    ('- foo: {{ bar }}', 'This may be an issue with missing quotes around a template block.', True, 1, 8),
    (' -  - {{ bar }}', 'This may be an issue with missing quotes around a template block.', True, 1, 7),
    (' -  - foo: {{ bar }}', 'This may be an issue with missing quotes around a template block.', True, 1, 12),

    ('"foo" foo', 'Values starting with a quote must end with the same quote.', True, 1, 1),
    ('foo: "foo" foo', 'Values starting with a quote must end with the same quote.', True, 1, 6),
    ('- "foo" foo', 'Values starting with a quote must end with the same quote.', True, 1, 3),
    ('- foo: "foo" foo', 'Values starting with a quote must end with the same quote.', True, 1, 8),
    (' -  - "foo" foo', 'Values starting with a quote must end with the same quote.', True, 1, 7),
    (' -  - foo: "foo" foo', 'Values starting with a quote must end with the same quote.', True, 1, 12),

    ('"foo" "foo"', 'Values starting with a quote must end with the same quote, and not contain that quote.', True, 1, 1),
    ('foo: "foo" "foo"', 'Values starting with a quote must end with the same quote, and not contain that quote.', True, 1, 6),
    ('- "foo" "foo"', 'Values starting with a quote must end with the same quote, and not contain that quote.', True, 1, 3),
    ('- foo: "foo" "foo"', 'Values starting with a quote must end with the same quote, and not contain that quote.', True, 1, 8),
    (' -  - "foo" "foo"', 'Values starting with a quote must end with the same quote, and not contain that quote.', True, 1, 7),
    (' -  - foo: "foo" "foo"', 'Values starting with a quote must end with the same quote, and not contain that quote.', True, 1, 12),

    # cases without list/dict handling

    ('aaa: bbb:', 'Colons in unquoted values must be followed by a non-space character.', True, 1, 9),
    ('aaa: bbb: ccc', 'Colons in unquoted values must be followed by a non-space character.', True, 1, 9),
    ('''- value == "x: 'x' x"''', 'Colons in unquoted values must be followed by a non-space character.', True, 1, 14),

    ('!!map 1\t2\t3', 'Tabs are usually invalid in YAML.', False, 1, 8),
    ('{a: 1, a: 1}', "Found duplicate mapping key 'a'.", False, 1, 8),

    (1, 'a string or stream input is required', False, None, None),  # wrong type (misuse of API)

    # cases where the underling MarkedYAMLError is used

    ('k1: v1\n  k2: v2', 'Mapping values are not allowed in this context.', False, 2, 5),
    ('k1: v1\n  k2: ":"', 'Mapping values are not allowed in this context.', False, 2, 5),

    (':', 'While parsing a block mapping did not find expected key.', False, 1, 1),
    ('!!map 1', 'Expected a mapping node, but found scalar.', False, 1, 1),
    ('[]: bad', 'While constructing a mapping found unhashable key.', False, 1, 1),
    ('{}: bad', 'While constructing a mapping found unhashable key.', False, 1, 1),
    ('"\\"', 'While scanning a quoted scalar found unexpected end of stream.', False, 1, 4),

    # DTFIX-FUTURE: add tests that use comments
))
def test_yaml_parser_error(
        content: t.Any,
        expected_message: str,
        expect_help_text: bool,
        line: int | None,
        col: int | None,
        mocker: pytest_mock.MockerFixture,
) -> None:
    set_duplicate_yaml_dict_key_config(mocker, 'error')

    expected_message = f'YAML parsing failed: {expected_message}'

    with tempfile.TemporaryDirectory() as tempdir:
        source_path = pathlib.Path(tempdir) / 'source.yml'
        source_path.write_text(str(content))

        with pytest.raises(AnsibleYAMLParserError) as error:
            from_yaml(content, file_name=str(source_path))

    assert error.value.message == expected_message
    assert error.value._original_message == expected_message
    assert format_exception_message(error.value) == expected_message
    assert str(error.value) == expected_message

    assert error.value.obj == Origin(path=str(source_path), line_num=line, col_num=col)

    if expect_help_text:
        assert error.value._help_text is not None  # DTFIX-FUTURE: check the content later once it's less volatile
    else:
        assert error.value._help_text is None


def test_yaml_duplicate_key_warning(mocker: pytest_mock.MockerFixture) -> None:
    set_duplicate_yaml_dict_key_config(mocker, 'warn')

    patched_warning = mocker.patch.object(Display(), 'warning')

    assert from_yaml('{a: 1, b: 2, c: 3, b: 4, d: 5, b: 6}', file_name='/hello.yml')

    patched_warning.assert_has_calls((
        call(msg="Found duplicate mapping key 'b'.", obj="b", help_text="Using last defined value only."),
        call(msg="Found duplicate mapping key 'b'.", obj="b", help_text="Using last defined value only."),
    ))


def test_yaml_duplicate_key_ignore(mocker: pytest_mock.MockerFixture) -> None:
    set_duplicate_yaml_dict_key_config(mocker, 'ignore')

    warning_spy = mocker.spy(Display(), 'warning')

    assert from_yaml('{a: 1, a: 2}', file_name='/hello.yml')

    warning_spy.assert_not_called()


def set_duplicate_yaml_dict_key_config(mocker: pytest_mock.MockerFixture, value: str):
    original_get_config_value = C.config.get_config_value

    def mocked_get_config_value(config, *args, **kwargs):
        if config == 'DUPLICATE_YAML_DICT_KEY':
            return value

        return original_get_config_value(config, *args, **kwargs)

    mocker.patch.object(C.config, 'get_config_value', mocked_get_config_value)



# Copyright (c) 2023 Ansible
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import annotations

import datetime as _datetime
import typing as t

from ansible.module_utils.common.warnings import deprecate


_UTC = _datetime.timezone.utc


def _utcfromtimestamp(timestamp: float) -> _datetime.datetime:
    """Construct an aware UTC datetime from a POSIX timestamp."""
    return _datetime.datetime.fromtimestamp(timestamp, _UTC)


def _utcnow() -> _datetime.datetime:
    """Construct an aware UTC datetime from time.time()."""
    return _datetime.datetime.now(_UTC)


_deprecated_shims_map: dict[str, t.Callable[..., object] | _datetime.timezone] = {
    'UTC': _UTC,
    'utcfromtimestamp': _utcfromtimestamp,
    'utcnow': _utcnow,
}

__all__ = tuple(_deprecated_shims_map)


def __getattr__(importable_name: str) -> t.Callable[..., object] | _datetime.timezone:
    """Inject import-time deprecation warnings.

    Specifically, for ``UTC``, ``utcfromtimestamp()`` and ``utcnow()``.
    """
    try:
        importable = _deprecated_shims_map[importable_name]
    except KeyError as key_err:
        raise AttributeError(f"module {__name__!r} has no attribute {key_err}") from None

    deprecate(
        msg=f'The `ansible.module_utils.compat.datetime.{importable_name}` '
        'function is deprecated.',
        version='2.21',
    )

    return importable



"""
Remote debugging support for AnsiballZ modules with debugpy.

To use with VS Code:

1) Choose an available port for VS Code to listen on (e.g. 5678).
2) Ensure `debugpy` is installed for the interpreter(s) which will run the code being debugged.
3) Create the following launch.json configuration

    {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "Python Debug Server",
                "type": "debugpy",
                "request": "attach",
                "listen": {
                    "host": "localhost",
                    "port": 5678,
                },
            },
            {
                "name": "ansible-playbook main.yml",
                "type": "debugpy",
                "request": "launch",
                "module": "ansible",
                "args": [
                    "playbook",
                    "main.yml"
                ],
                "env": {
                    "_ANSIBLE_ANSIBALLZ_DEBUGPY_CONFIG": "{\"host\": \"localhost\", \"port\": 5678}"
                },
                "console": "integratedTerminal",
            }
        ],
        "compounds": [
            {
                "name": "Test Module Debugging",
                "configurations": [
                    "Python Debug Server",
                    "ansible-playbook main.yml"
                ],
                "stopAll": true
            }
        ]
    }

4) Set any desired breakpoints.
5) Configure the Run and Debug view to use the "Test Module Debugging" compound configuration.
6) Press F5 to start debugging.
"""

from __future__ import annotations

import dataclasses
import json
import os
import pathlib

import typing as t


@dataclasses.dataclass(frozen=True)
class Options:
    """Debugger options for debugpy."""

    host: str = 'localhost'
    """The host to connect to for remote debugging."""
    port: int = 5678
    """The port to connect to for remote debugging."""
    connect: dict[str, object] = dataclasses.field(default_factory=dict)
    """The options to pass to the `debugpy.connect` method."""
    source_mapping: dict[str, str] = dataclasses.field(default_factory=dict)
    """
    A mapping of source paths to provide to debugpy.
    This setting is used internally by AnsiballZ and is not required unless Ansible CLI commands are run from a different system than your IDE.
    In that scenario, use this setting instead of configuring source mapping in your IDE.
    The key is a path known to the IDE.
    The value is the same path as known to the Ansible CLI.
    Both file paths and directories are supported.
    """


def run(args: dict[str, t.Any]) -> None:  # pragma: nocover
    """Enable remote debugging."""
    import debugpy

    options = Options(**args)
    temp_dir = pathlib.Path(__file__).parent.parent.parent.parent.parent.parent
    path_mapping = [[key, str(temp_dir / value)] for key, value in options.source_mapping.items()]

    os.environ['PATHS_FROM_ECLIPSE_TO_PYTHON'] = json.dumps(path_mapping)

    debugpy.connect((options.host, options.port), **options.connect)

    pass  # A convenient place to put a breakpoint



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pytest

from ansible.module_utils.facts.system.fips import FipsFactCollector


@pytest.mark.parametrize(("return_value", "expected"), [('1', True), ('0', False)])
def test_fips(mocker, return_value, expected):
    mocker.patch('ansible.module_utils.facts.system.fips.get_file_content', return_value=return_value)
    fips_mgr = FipsFactCollector().collect()
    assert fips_mgr['fips'] is expected



# -*- coding: utf-8 -*-

# Copyright: (c) 2012, Dag Wieers <dag@wieers.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = r"""
---
module: assert
short_description: Asserts given expressions are true
description:
     - This module asserts that given expressions are true with an optional custom message.
     - This module is also supported for Windows targets.
version_added: "1.5"
options:
  that:
    description:
      - A list of string expressions of the same form that can be passed to the C(when) statement.
    type: list
    elements: str
    required: true
  fail_msg:
    description:
      - The customized message used for a failing assertion.
      - This argument was called O(msg) before Ansible 2.7, now it is renamed to O(fail_msg) with alias O(msg).
    type: str
    aliases: [ msg ]
    version_added: "2.7"
  success_msg:
    description:
      - The customized message used for a successful assertion.
    type: str
    version_added: "2.7"
  quiet:
    description:
      - Set this to V(true) to avoid verbose output.
    type: bool
    default: no
    version_added: "2.8"
extends_documentation_fragment:
  - action_common_attributes
  - action_common_attributes.conn
  - action_common_attributes.flow
attributes:
    action:
        support: full
    async:
        support: none
    become:
        support: none
    bypass_host_loop:
        support: none
    connection:
        support: none
    check_mode:
        support: full
    delegation:
        support: none
        details: Aside from C(register) and/or in combination with C(delegate_facts), it has little effect.
    diff_mode:
        support: none
    platform:
        platforms: all
seealso:
- module: ansible.builtin.debug
- module: ansible.builtin.fail
- module: ansible.builtin.meta
author:
    - Ansible Core Team
    - Michael DeHaan
"""

EXAMPLES = r"""
- name: A single condition can be supplied as string instead of list
  ansible.builtin.assert:
    that: "ansible_os_family != 'RedHat'"

- name: Use yaml multiline strings to ease escaping
  ansible.builtin.assert:
    that:
      - "'foo' in some_command_result.stdout"
      - number_of_the_counting == 3
      - >
        "reject" not in some_command_result.stderr

- name: After version 2.7 both O(msg) and O(fail_msg) can customize failing assertion message
  ansible.builtin.assert:
    that:
      - my_param <= 100
      - my_param >= 0
    fail_msg: "'my_param' must be between 0 and 100"
    success_msg: "'my_param' is between 0 and 100"

- name: Please use O(msg) when ansible version is smaller than 2.7
  ansible.builtin.assert:
    that:
      - my_param <= 100
      - my_param >= 0
    msg: "'my_param' must be between 0 and 100"

- name: Use quiet to avoid verbose output
  ansible.builtin.assert:
    that:
      - my_param <= 100
      - my_param >= 0
    quiet: true
"""



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule


class Die(BaseException):
    """Raise an exception that is unlikely to be caught (and will thus catastrophically exit the process)."""


def main():
    module = AnsibleModule(
        argument_spec=dict(
            errors=dict(type='list', elements='str', default=[]),
            trigger_unhandled_exception=dict(type='str', default=''),
            return_from_main=dict(type='bool', default=False),
        ),
        supports_check_mode=True,
        mutually_exclusive=[['errors', 'trigger_unhandled_exception', 'return_from_main']],
    )

    if module.params['return_from_main']:
        return

    if trigger_unhandled_exception := module.params['trigger_unhandled_exception']:
        raise Die(trigger_unhandled_exception)

    if errors := module.params['errors']:
        raise_exceptions(errors)

    result = dict()

    module.exit_json(**result)


def raise_exceptions(exceptions: list[str]) -> None:
    if len(exceptions) > 1:
        try:
            raise_exceptions(exceptions[1:])
        except Exception as ex:
            raise Exception(exceptions[0]) from ex

    raise Exception(exceptions[0])


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import pytest

from ansible.modules.get_url import parse_digest_lines

FILENAME = "sample.txt"


@pytest.mark.parametrize(
    ("lines", "expected"),
    [
        pytest.param(
            [
                "2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25 sample.txt",
            ],
            [("2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25", FILENAME)],
            id="single-line-digest-single-space",
        ),
        pytest.param(
            [
                "2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25         sample.txt",
            ],
            [("2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25", FILENAME)],
            id="single-line-digest-multiple-spaces",
        ),
        pytest.param(
            [
                "2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25         .sample.txt",
            ],
            [("2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25", FILENAME)],
            id="single-line-digest-multiple-spaces-with-dot",
        ),
        pytest.param(
            [
                "2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25         *sample.txt",
            ],
            [("2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25", FILENAME)],
            id="single-line-digest-multiple-spaces-with-asterisk",
        ),
        pytest.param(
            [
                "2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25         ./sample.txt",
            ],
            [("2a32d433bf82355a3f78318a5affa21866c9a98b151785494b386e6b08f40b25", FILENAME)],
            id="single-line-digest-multiple-spaces-with-dot-and-slash",
        ),
        pytest.param(
            [
                "a97e6837f60cec6da4491bab387296bbcd72bdba",
            ],
            [("a97e6837f60cec6da4491bab387296bbcd72bdba", FILENAME)],
            id="single-line-digest",
        ),
        pytest.param(
            [
                "a97e6837f60cec6da4491bab387296bbcd72bdba  sample.txt",
            ],
            [("a97e6837f60cec6da4491bab387296bbcd72bdba", FILENAME)],
            id="GNU-style-digest",
        ),
        pytest.param(
            [
                "SHA256 (sample.txt) = b1b6ce5073c8fac263a8fc5edfffdbd5dec1980c784e09c5bc69f8fb6056f006.",
            ],
            [
                (
                    "b1b6ce5073c8fac263a8fc5edfffdbd5dec1980c784e09c5bc69f8fb6056f006.",
                    FILENAME,
                )
            ],
            id="BSD-style-digest",
        ),
    ],
)
def test_parse_digest_lines(lines, expected):
    assert parse_digest_lines(filename=FILENAME, lines=lines) == expected



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os
import pytest

from ansible.module_utils.facts.system.distribution import DistributionFiles


@pytest.mark.parametrize(
    ('distro_file', 'expected_version'),
    (
        ('Slackware', '14.1'),
        ('SlackwareCurrent', '14.2+'),
    )
)
def test_parse_distribution_file_slackware(mock_module, distro_file, expected_version):
    with open(os.path.join(os.path.dirname(__file__), '../../fixtures/distribution_files', distro_file)) as file:
        data = file.read()

    test_input = {
        'name': 'Slackware',
        'data': data,
        'path': '/etc/os-release',
        'collected_facts': None,
    }

    result = (
        True,
        {
            'distribution': 'Slackware',
            'distribution_version': expected_version
        }
    )
    distribution = DistributionFiles(module=mock_module())
    assert result == distribution.parse_distribution_file_Slackware(**test_input)



from __future__ import annotations

import re


def test_name_ok(value):
    # test names include a unique hash value to prevent shadowing of other plugins
    return bool(re.match(r'^ansible\.plugins\.test\.test_test_[0-9]+$', __name__))


class TestModule:
    def tests(self):
        return {
            'test_name_ok': test_name_ok,
        }



from __future__ import annotations

import copy
import pickle

import pytest
import jinja2
import jinja2.utils

from ansible._internal._templating._jinja_common import MarkerError, Marker


# NB: this module is named with a trailing underscore to work around a PyCharm/pydevd bug:
#  https://youtrack.jetbrains.com/issue/PY-70408/Debugger-skips-all-breakpoints-in-a-single-test-file


def get_dunder_methods_to_intercept() -> list[str]:
    """
    Return a list of dunder methods on Jinja's StrictUndefined that should be overridden by Ansible's Marker.
    When new methods are added in future Python/Jinja versions, they need to be added to Marker or the ignore_names list below.
    """
    dunder_names = set(name for name in dir(jinja2.StrictUndefined) if name.startswith('__') and name.endswith('__'))

    strict_undefined_intercepted_method_names = set(
        name for name in dir(jinja2.StrictUndefined)
        if getattr(jinja2.StrictUndefined, name) is jinja2.StrictUndefined._fail_with_undefined_error and name != '_fail_with_undefined_error'
    )

    # Some attributes/methods are necessary for core Python interactions and must not be intercepted, thus they are excluded from this test.
    ignore_names = {
        '__class__',
        '__dir__',
        '__doc__',
        '__firstlineno__',
        '__getattr__',  # tested separately since it is intercepted with a custom method
        '__getattribute__',
        '__getitem__',  # tested separately since it is intercepted with a custom method
        '__getstate__',
        '__init__',
        '__init_subclass__',
        '__module__',
        '__new__',
        '__reduce__',
        '__reduce_ex__',
        '__setattr__',  # tested separately since it is intercepted with a custom method
        '__slots__',  # tested separately since it's not a method
        '__sizeof__',
        '__static_attributes__',
        '__subclasshook__',
    }

    # Some methods not intercepted by Jinja's StrictUndefined should be intercepted by Marker.
    additional_method_names = {
        '__aiter__',
        '__delattr__',
        '__format__',
        '__repr__',
        '__setitem__',
    }

    assert not strict_undefined_intercepted_method_names - dunder_names  # ensure Jinja intercepted methods have not been overlooked
    assert not ignore_names & additional_method_names  # ensure no overap between ignore_names and additional_method_names

    return sorted(dunder_names - ignore_names | additional_method_names)


def test_jinja_undefined_shape():
    """
    Assert that the internal shape of Jinja's StrictUndefined matches Marker's expectations.
    If this test fails, it likely means Jinja has changed something about the internals of Undefined in a way we'll need to address.
    """
    assert jinja2.StrictUndefined in Marker.__bases__  # ensure we're directly inheriting the base we're validating

    required_attrs = (
        '_undefined_message',
        '_undefined_name',
        '_fail_with_undefined_error',
    )

    assert all(hasattr(jinja2.StrictUndefined, a) for a in required_attrs)

    assert isinstance(jinja2.StrictUndefined.__slots__, tuple)  # we don't actually care what's slotted, just that the attr exists (or has been patched back on)


@pytest.mark.parametrize("name", get_dunder_methods_to_intercept())
def test_marker_methods(marker: Marker, name: str) -> None:
    """Verify all expected dunder methods on Marker raise a MarkerError."""
    method = getattr(marker, name)

    with pytest.raises(MarkerError) as err:
        method()

    assert err.value.source is marker


def test_getattr_attribute_error(marker: Marker) -> None:
    """Verify unknown dunder attributes on Marker raise an AttributeError."""
    with pytest.raises(AttributeError) as err:
        getattr(marker, '__does_not_exist__')

    assert err.value.obj is marker
    assert err.value.name == '__does_not_exist__'


def test_getattr_propagation(marker: Marker) -> None:
    """Verify unknown non-dunder attributes on Marker return self."""
    assert marker.does_not_exist is marker


def test_getitem_propagation(marker: Marker) -> None:
    """Verify items return self."""
    assert marker['does_not_exist'] is marker


def test_setattr(marker: Marker) -> None:
    """Verify setattr raises MarkerError."""
    with pytest.raises(MarkerError) as err:
        marker.something = True

    assert err.value.source is marker


def test_setitem(marker: Marker) -> None:
    """Verify setitem raises MarkerError."""
    with pytest.raises(MarkerError) as err:
        marker['something'] = True

    assert err.value.source is marker


def test_slots(marker: Marker) -> None:
    """Verify all Marker instances are slotted."""
    assert isinstance(marker.__slots__, tuple)
    assert not hasattr(marker, '__dict__')


def marker_attrs(marker: Marker) -> list[str]:
    """Returns a list of internal attributes for a Marker type (using known prefixes)."""
    return [name for name in dir(marker) if name.startswith('_undefined_') or name.startswith('_marker_')]


def test_copy(marker: Marker) -> None:
    """Verify copying an Marker works."""
    copied = copy.copy(marker)

    assert copied is not marker

    for attribute_name in marker_attrs(marker):
        assert getattr(copied, attribute_name) is getattr(marker, attribute_name), attribute_name


def test_deepcopy(marker: Marker) -> None:
    """Verify deep copying an Marker works."""
    copied = copy.deepcopy(marker)

    assert copied is not marker

    for attribute_name in marker_attrs(marker):
        copied_value = getattr(copied, attribute_name)
        marker_value = getattr(marker, attribute_name)

        if attribute_name == '_undefined_exception':
            # The `_undefined_exception` attribute is a type, so the identity remains unchanged.
            assert copied_value is marker_value, attribute_name
        elif attribute_name == '_undefined_obj' and marker_value is jinja2.utils.missing:
            # The `_undefined_obj` attribute defaults to the `jinja2.utils.missing`, a singleton of `jinja2.utils.MissingType`.
            assert copied_value is marker_value, attribute_name
        elif type(marker_value) in (str, type(None)):
            # Values of type `str` or `None` should be equal.
            assert copied_value == marker_value, attribute_name
        else:
            # All other types should be actual copies.
            assert copied_value is not marker_value, attribute_name


def test_pickle(marker: Marker) -> None:
    """Verify pickling a Marker works."""
    pickled = pickle.loads(pickle.dumps(marker))

    assert pickled is not marker

    for attribute_name in marker_attrs(marker):
        pickled_value = getattr(pickled, attribute_name)
        marker_value = getattr(marker, attribute_name)

        if attribute_name == '_marker_captured_exception':
            # The `_marker_captured_exception` attribute is an `Exception` instance, which doesn't support equality comparisons, compare as a `str` instead.
            pickled_value = str(pickled_value)
            marker_value = str(marker_value)

        assert pickled_value == marker_value, attribute_name



from __future__ import annotations

import concurrent.futures as _cf
import subprocess
import sys
import time

import pytest

from ansible.module_utils._internal._concurrent import _futures


def test_daemon_thread_pool_nonblocking_cm_exit() -> None:
    """Ensure that the ThreadPoolExecutor context manager exit is not blocked by in-flight tasks."""
    with _futures.DaemonThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(time.sleep, 5)

        with pytest.raises(_cf.TimeoutError):  # deprecated: description='aliased to stdlib TimeoutError in 3.11' python_version='3.10'
            future.result(timeout=1)

    assert future.running()  # ensure the future is still going (ie, we didn't have to wait for it to return)


_task_success_msg = "work completed"
_process_success_msg = "exit success"
_timeout_sec = 3
_sleep_time_sec = _timeout_sec * 2


def test_blocking_shutdown() -> None:
    """Run with the DaemonThreadPoolExecutor patch disabled to verify that shutdown is blocked by in-flight tasks."""
    with pytest.raises(subprocess.TimeoutExpired):
        subprocess.run(args=[sys.executable, __file__, 'block'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=True, timeout=_timeout_sec)


def test_non_blocking_shutdown() -> None:
    """Run with the DaemonThreadPoolExecutor patch enabled to verify that shutdown is not blocked by in-flight tasks."""
    cp = subprocess.run(args=[sys.executable, __file__, ''], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=True, timeout=_timeout_sec)

    assert _task_success_msg in cp.stdout
    assert _process_success_msg in cp.stdout


def _run_blocking_exit_test(use_patched: bool) -> None:  # pragma: nocover
    """Helper for external process integration test."""
    tpe_type = _futures.DaemonThreadPoolExecutor if use_patched else _cf.ThreadPoolExecutor

    with tpe_type(max_workers=2) as tp:
        fs_non_blocking = tp.submit(lambda: print(_task_success_msg))
        assert [tp.submit(time.sleep, _sleep_time_sec) for _idx in range(4)]  # not a pointless statement
        fs_non_blocking.result(timeout=1)

    print(_process_success_msg)


def main() -> None:  # pragma: nocover
    """Used by test_(non)blocking_shutdown as a script-style run."""
    _run_blocking_exit_test(sys.argv[1] != 'block')


if __name__ == '__main__':  # pragma: nocover
    main()



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule
from ..module_utils.shared_deprecation import get_deprecation_kwargs, get_deprecated_value


def main() -> None:
    m = AnsibleModule({})
    m.warn("This is a warning.")

    for deprecate_kw in get_deprecation_kwargs():
        m.deprecate(**deprecate_kw)  # pylint: disable=ansible-deprecated-no-version

    m.exit_json(deprecated_result=get_deprecated_value())


if __name__ == '__main__':
    main()



# Copyright: Contributors to the Ansible project
# BSD 3 Clause License (see licenses/BSD-3-Clause.txt or https://opensource.org/license/bsd-3-clause/)

from __future__ import annotations

import binascii
import copy
import dataclasses
import enum
import functools
import hashlib
import socket
import types
import typing as t

try:
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric.dsa import (
        DSAParameterNumbers,
        DSAPrivateKey,
        DSAPublicKey,
        DSAPublicNumbers,
    )
    from cryptography.hazmat.primitives.asymmetric.ec import (
        EllipticCurve,
        EllipticCurvePrivateKey,
        EllipticCurvePublicKey,
        SECP256R1,
        SECP384R1,
        SECP521R1,
    )
    from cryptography.hazmat.primitives.asymmetric.ed25519 import (
        Ed25519PrivateKey,
        Ed25519PublicKey,
    )
    from cryptography.hazmat.primitives.asymmetric.rsa import (
        RSAPrivateKey,
        RSAPublicKey,
        RSAPublicNumbers,
    )
except ImportError:
    HAS_CRYPTOGRAPHY = False
else:
    HAS_CRYPTOGRAPHY = True

    CryptoPublicKey = t.Union[
        DSAPublicKey,
        EllipticCurvePublicKey,
        Ed25519PublicKey,
        RSAPublicKey,
    ]

    CryptoPrivateKey = t.Union[
        DSAPrivateKey,
        EllipticCurvePrivateKey,
        Ed25519PrivateKey,
        RSAPrivateKey,
    ]


if t.TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric.dsa import DSAPrivateNumbers
    from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateNumbers
    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateNumbers


_SSH_AGENT_CLIENT_SOCKET_TIMEOUT = 10


class ProtocolMsgNumbers(enum.IntEnum):
    # Responses
    SSH_AGENT_FAILURE = 5
    SSH_AGENT_SUCCESS = 6
    SSH_AGENT_IDENTITIES_ANSWER = 12
    SSH_AGENT_SIGN_RESPONSE = 14
    SSH_AGENT_EXTENSION_FAILURE = 28
    SSH_AGENT_EXTENSION_RESPONSE = 29

    # Constraints
    SSH_AGENT_CONSTRAIN_LIFETIME = 1
    SSH_AGENT_CONSTRAIN_CONFIRM = 2
    SSH_AGENT_CONSTRAIN_EXTENSION = 255

    # Requests
    SSH_AGENTC_REQUEST_IDENTITIES = 11
    SSH_AGENTC_SIGN_REQUEST = 13
    SSH_AGENTC_ADD_IDENTITY = 17
    SSH_AGENTC_REMOVE_IDENTITY = 18
    SSH_AGENTC_REMOVE_ALL_IDENTITIES = 19
    SSH_AGENTC_ADD_SMARTCARD_KEY = 20
    SSH_AGENTC_REMOVE_SMARTCARD_KEY = 21
    SSH_AGENTC_LOCK = 22
    SSH_AGENTC_UNLOCK = 23
    SSH_AGENTC_ADD_ID_CONSTRAINED = 25
    SSH_AGENTC_ADD_SMARTCARD_KEY_CONSTRAINED = 26
    SSH_AGENTC_EXTENSION = 27

    def to_blob(self) -> bytes:
        return bytes([self])


class SshAgentFailure(RuntimeError):
    """Server failure or unexpected response."""


# NOTE: Classes below somewhat represent "Data Type Representations Used in the SSH Protocols"
#       as specified by RFC4251


@t.runtime_checkable
class SupportsToBlob(t.Protocol):
    def to_blob(self) -> bytes: ...


@t.runtime_checkable
class SupportsFromBlob(t.Protocol):
    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self: ...

    @classmethod
    def consume_from_blob(cls, blob: memoryview | bytes) -> tuple[t.Self, memoryview | bytes]: ...


def _split_blob(blob: memoryview | bytes, length: int) -> tuple[memoryview | bytes, memoryview | bytes]:
    if len(blob) < length:
        raise ValueError("_split_blob: unexpected data length")
    return blob[:length], blob[length:]


class VariableSized:
    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        raise NotImplementedError

    @classmethod
    def consume_from_blob(cls, blob: memoryview | bytes) -> tuple[t.Self, memoryview | bytes]:
        length = uint32.from_blob(blob[:4])
        blob = blob[4:]
        data, rest = _split_blob(blob, length)
        return cls.from_blob(data), rest


class uint32(int):
    def to_blob(self) -> bytes:
        return self.to_bytes(length=4, byteorder='big')

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        return cls.from_bytes(blob, byteorder='big')

    @classmethod
    def consume_from_blob(cls, blob: memoryview | bytes) -> tuple[t.Self, memoryview | bytes]:
        length = uint32(4)
        data, rest = _split_blob(blob, length)
        return cls.from_blob(data), rest


class mpint(int, VariableSized):
    def to_blob(self) -> bytes:
        if self < 0:
            raise ValueError("negative mpint not allowed")
        if not self:
            return b""
        nbytes = (self.bit_length() + 8) // 8
        ret = bytearray(self.to_bytes(length=nbytes, byteorder='big'))
        ret[:0] = uint32(len(ret)).to_blob()
        return ret

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        if blob and blob[0] > 127:
            raise ValueError("Invalid data")
        return cls.from_bytes(blob, byteorder='big')


class constraints(bytes):
    def to_blob(self) -> bytes:
        return self


class binary_string(bytes, VariableSized):
    def to_blob(self) -> bytes:
        if length := len(self):
            return uint32(length).to_blob() + self
        else:
            return b""

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        return cls(blob)


class unicode_string(str, VariableSized):
    def to_blob(self) -> bytes:
        val = self.encode('utf-8')
        if length := len(val):
            return uint32(length).to_blob() + val
        else:
            return b""

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        return cls(bytes(blob).decode('utf-8'))


class KeyAlgo(str, VariableSized, enum.Enum):
    RSA = "ssh-rsa"
    DSA = "ssh-dss"
    ECDSA256 = "ecdsa-sha2-nistp256"
    SKECDSA256 = "sk-ecdsa-sha2-nistp256@openssh.com"
    ECDSA384 = "ecdsa-sha2-nistp384"
    ECDSA521 = "ecdsa-sha2-nistp521"
    ED25519 = "ssh-ed25519"
    SKED25519 = "sk-ssh-ed25519@openssh.com"
    RSASHA256 = "rsa-sha2-256"
    RSASHA512 = "rsa-sha2-512"

    @property
    def main_type(self) -> str:
        match self:
            case self.RSA:
                return 'RSA'
            case self.DSA:
                return 'DSA'
            case self.ECDSA256 | self.ECDSA384 | self.ECDSA521:
                return 'ECDSA'
            case self.ED25519:
                return 'ED25519'
            case _:
                raise NotImplementedError(self.name)

    def to_blob(self) -> bytes:
        b_self = self.encode('utf-8')
        return uint32(len(b_self)).to_blob() + b_self

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        return cls(bytes(blob).decode('utf-8'))


if HAS_CRYPTOGRAPHY:
    _ECDSA_KEY_TYPE: dict[KeyAlgo, type[EllipticCurve]] = {
        KeyAlgo.ECDSA256: SECP256R1,
        KeyAlgo.ECDSA384: SECP384R1,
        KeyAlgo.ECDSA521: SECP521R1,
    }


@dataclasses.dataclass
class Msg:
    def to_blob(self) -> bytes:
        rv = bytearray()
        for field in dataclasses.fields(self):
            fv = getattr(self, field.name)
            if isinstance(fv, SupportsToBlob):
                rv.extend(fv.to_blob())
            else:
                raise NotImplementedError(field.type)
        return rv

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self:
        args: list[t.Any] = []
        for _field_name, field_type in t.get_type_hints(cls).items():
            if isinstance(field_type, SupportsFromBlob):
                fv, blob = field_type.consume_from_blob(blob)
                args.append(fv)
            else:
                raise NotImplementedError(str(field_type))
        return cls(*args)


@dataclasses.dataclass
class PrivateKeyMsg(Msg):
    @staticmethod
    def from_private_key(private_key: CryptoPrivateKey) -> PrivateKeyMsg:
        match private_key:
            case RSAPrivateKey():
                rsa_pn: RSAPrivateNumbers = private_key.private_numbers()
                return RSAPrivateKeyMsg(
                    KeyAlgo.RSA,
                    mpint(rsa_pn.public_numbers.n),
                    mpint(rsa_pn.public_numbers.e),
                    mpint(rsa_pn.d),
                    mpint(rsa_pn.iqmp),
                    mpint(rsa_pn.p),
                    mpint(rsa_pn.q),
                )
            case DSAPrivateKey():
                dsa_pn: DSAPrivateNumbers = private_key.private_numbers()
                return DSAPrivateKeyMsg(
                    KeyAlgo.DSA,
                    mpint(dsa_pn.public_numbers.parameter_numbers.p),
                    mpint(dsa_pn.public_numbers.parameter_numbers.q),
                    mpint(dsa_pn.public_numbers.parameter_numbers.g),
                    mpint(dsa_pn.public_numbers.y),
                    mpint(dsa_pn.x),
                )
            case EllipticCurvePrivateKey():
                ecdsa_pn: EllipticCurvePrivateNumbers = private_key.private_numbers()
                key_size = private_key.key_size
                return EcdsaPrivateKeyMsg(
                    getattr(KeyAlgo, f'ECDSA{key_size}'),
                    unicode_string(f'nistp{key_size}'),
                    binary_string(
                        private_key.public_key().public_bytes(
                            encoding=serialization.Encoding.X962,
                            format=serialization.PublicFormat.UncompressedPoint,
                        )
                    ),
                    mpint(ecdsa_pn.private_value),
                )
            case Ed25519PrivateKey():
                public_bytes = private_key.public_key().public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw,
                )
                private_bytes = private_key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption(),
                )
                return Ed25519PrivateKeyMsg(
                    KeyAlgo.ED25519,
                    binary_string(public_bytes),
                    binary_string(private_bytes + public_bytes),
                )
            case _:
                raise NotImplementedError(private_key)


@dataclasses.dataclass(order=True, slots=True)
class RSAPrivateKeyMsg(PrivateKeyMsg):
    type: KeyAlgo
    n: mpint
    e: mpint
    d: mpint
    iqmp: mpint
    p: mpint
    q: mpint
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)
    constraints: constraints = dataclasses.field(default=constraints(b''))


@dataclasses.dataclass(order=True, slots=True)
class DSAPrivateKeyMsg(PrivateKeyMsg):
    type: KeyAlgo
    p: mpint
    q: mpint
    g: mpint
    y: mpint
    x: mpint
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)
    constraints: constraints = dataclasses.field(default=constraints(b''))


@dataclasses.dataclass(order=True, slots=True)
class EcdsaPrivateKeyMsg(PrivateKeyMsg):
    type: KeyAlgo
    ecdsa_curve_name: unicode_string
    Q: binary_string
    d: mpint
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)
    constraints: constraints = dataclasses.field(default=constraints(b''))


@dataclasses.dataclass(order=True, slots=True)
class Ed25519PrivateKeyMsg(PrivateKeyMsg):
    type: KeyAlgo
    enc_a: binary_string
    k_env_a: binary_string
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)
    constraints: constraints = dataclasses.field(default=constraints(b''))


@dataclasses.dataclass
class PublicKeyMsg(Msg):
    @staticmethod
    def get_dataclass(type: KeyAlgo) -> type[
        t.Union[
            RSAPublicKeyMsg,
            EcdsaPublicKeyMsg,
            Ed25519PublicKeyMsg,
            DSAPublicKeyMsg,
        ]
    ]:
        match type:
            case KeyAlgo.RSA:
                return RSAPublicKeyMsg
            case KeyAlgo.ECDSA256 | KeyAlgo.ECDSA384 | KeyAlgo.ECDSA521:
                return EcdsaPublicKeyMsg
            case KeyAlgo.ED25519:
                return Ed25519PublicKeyMsg
            case KeyAlgo.DSA:
                return DSAPublicKeyMsg
            case _:
                raise NotImplementedError(type)

    @functools.cached_property
    def public_key(self) -> CryptoPublicKey:
        type: KeyAlgo = self.type
        match type:
            case KeyAlgo.RSA:
                return RSAPublicNumbers(self.e, self.n).public_key()
            case KeyAlgo.ECDSA256 | KeyAlgo.ECDSA384 | KeyAlgo.ECDSA521:
                curve = _ECDSA_KEY_TYPE[KeyAlgo(type)]
                return EllipticCurvePublicKey.from_encoded_point(curve(), self.Q)
            case KeyAlgo.ED25519:
                return Ed25519PublicKey.from_public_bytes(self.enc_a)
            case KeyAlgo.DSA:
                return DSAPublicNumbers(self.y, DSAParameterNumbers(self.p, self.q, self.g)).public_key()
            case _:
                raise NotImplementedError(type)

    @staticmethod
    def from_public_key(public_key: CryptoPublicKey) -> PublicKeyMsg:
        match public_key:
            case DSAPublicKey():
                dsa_pn: DSAPublicNumbers = public_key.public_numbers()
                return DSAPublicKeyMsg(
                    KeyAlgo.DSA,
                    mpint(dsa_pn.parameter_numbers.p),
                    mpint(dsa_pn.parameter_numbers.q),
                    mpint(dsa_pn.parameter_numbers.g),
                    mpint(dsa_pn.y),
                )
            case EllipticCurvePublicKey():
                return EcdsaPublicKeyMsg(
                    getattr(KeyAlgo, f'ECDSA{public_key.curve.key_size}'),
                    unicode_string(f'nistp{public_key.curve.key_size}'),
                    binary_string(
                        public_key.public_bytes(
                            encoding=serialization.Encoding.X962,
                            format=serialization.PublicFormat.UncompressedPoint,
                        )
                    ),
                )
            case Ed25519PublicKey():
                return Ed25519PublicKeyMsg(
                    KeyAlgo.ED25519,
                    binary_string(
                        public_key.public_bytes(
                            encoding=serialization.Encoding.Raw,
                            format=serialization.PublicFormat.Raw,
                        )
                    ),
                )
            case RSAPublicKey():
                rsa_pn: RSAPublicNumbers = public_key.public_numbers()
                return RSAPublicKeyMsg(KeyAlgo.RSA, mpint(rsa_pn.e), mpint(rsa_pn.n))
            case _:
                raise NotImplementedError(public_key)

    @functools.cached_property
    def fingerprint(self) -> str:
        digest = hashlib.sha256()
        msg = copy.copy(self)
        msg.comments = unicode_string('')
        k = msg.to_blob()
        digest.update(k)
        return binascii.b2a_base64(digest.digest(), newline=False).rstrip(b'=').decode('utf-8')


@dataclasses.dataclass(order=True, slots=True)
class RSAPublicKeyMsg(PublicKeyMsg):
    type: KeyAlgo
    e: mpint
    n: mpint
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)


@dataclasses.dataclass(order=True, slots=True)
class DSAPublicKeyMsg(PublicKeyMsg):
    type: KeyAlgo
    p: mpint
    q: mpint
    g: mpint
    y: mpint
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)


@dataclasses.dataclass(order=True, slots=True)
class EcdsaPublicKeyMsg(PublicKeyMsg):
    type: KeyAlgo
    ecdsa_curve_name: unicode_string
    Q: binary_string
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)


@dataclasses.dataclass(order=True, slots=True)
class Ed25519PublicKeyMsg(PublicKeyMsg):
    type: KeyAlgo
    enc_a: binary_string
    comments: unicode_string = dataclasses.field(default=unicode_string(''), compare=False)


@dataclasses.dataclass(order=True, slots=True)
class KeyList(Msg):
    nkeys: uint32
    keys: PublicKeyMsgList

    def __post_init__(self) -> None:
        if self.nkeys != len(self.keys):
            raise SshAgentFailure("agent: invalid number of keys received for identities list")


@dataclasses.dataclass(order=True, slots=True)
class PublicKeyMsgList(Msg):
    keys: list[PublicKeyMsg]

    def __iter__(self) -> t.Iterator[PublicKeyMsg]:
        yield from self.keys

    def __len__(self) -> int:
        return len(self.keys)

    @classmethod
    def from_blob(cls, blob: memoryview | bytes) -> t.Self: ...

    @classmethod
    def consume_from_blob(cls, blob: memoryview | bytes) -> tuple[t.Self, memoryview | bytes]:
        args: list[PublicKeyMsg] = []
        while blob:
            prev_blob = blob
            key_blob, key_blob_length, comment_blob = cls._consume_field(blob)

            peek_key_algo, _length, _blob = cls._consume_field(key_blob)
            pub_key_msg_cls = PublicKeyMsg.get_dataclass(KeyAlgo(bytes(peek_key_algo).decode('utf-8')))

            _fv, comment_blob_length, blob = cls._consume_field(comment_blob)
            key_plus_comment = prev_blob[4 : (4 + key_blob_length) + (4 + comment_blob_length)]

            args.append(pub_key_msg_cls.from_blob(key_plus_comment))
        return cls(args), b""

    @staticmethod
    def _consume_field(blob: memoryview | bytes) -> tuple[memoryview | bytes, uint32, memoryview | bytes]:
        length = uint32.from_blob(blob[:4])
        blob = blob[4:]
        data, rest = _split_blob(blob, length)
        return data, length, rest


class SshAgentClient:
    def __init__(self, auth_sock: str) -> None:
        self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._sock.settimeout(_SSH_AGENT_CLIENT_SOCKET_TIMEOUT)
        self._sock.connect(auth_sock)

    def close(self) -> None:
        self._sock.close()

    def __enter__(self) -> t.Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.close()

    def send(self, msg: bytes) -> bytes:
        length = uint32(len(msg)).to_blob()
        self._sock.sendall(length + msg)
        bufsize = uint32.from_blob(self._sock.recv(4))
        resp = self._sock.recv(bufsize)
        if resp[0] == ProtocolMsgNumbers.SSH_AGENT_FAILURE:
            raise SshAgentFailure('agent: failure')
        return resp

    def remove_all(self) -> None:
        self.send(ProtocolMsgNumbers.SSH_AGENTC_REMOVE_ALL_IDENTITIES.to_blob())

    def remove(self, public_key: CryptoPublicKey) -> None:
        key_blob = PublicKeyMsg.from_public_key(public_key).to_blob()
        self.send(ProtocolMsgNumbers.SSH_AGENTC_REMOVE_IDENTITY.to_blob() + uint32(len(key_blob)).to_blob() + key_blob)

    def add(
        self,
        private_key: CryptoPrivateKey,
        comments: str | None = None,
        lifetime: int | None = None,
        confirm: bool | None = None,
    ) -> None:
        key_msg = PrivateKeyMsg.from_private_key(private_key)
        key_msg.comments = unicode_string(comments or '')
        if lifetime:
            key_msg.constraints += constraints([ProtocolMsgNumbers.SSH_AGENT_CONSTRAIN_LIFETIME]).to_blob() + uint32(lifetime).to_blob()
        if confirm:
            key_msg.constraints += constraints([ProtocolMsgNumbers.SSH_AGENT_CONSTRAIN_CONFIRM]).to_blob()

        if key_msg.constraints:
            msg = ProtocolMsgNumbers.SSH_AGENTC_ADD_ID_CONSTRAINED.to_blob()
        else:
            msg = ProtocolMsgNumbers.SSH_AGENTC_ADD_IDENTITY.to_blob()
        msg += key_msg.to_blob()
        self.send(msg)

    def list(self) -> KeyList:
        req = ProtocolMsgNumbers.SSH_AGENTC_REQUEST_IDENTITIES.to_blob()
        r = memoryview(bytearray(self.send(req)))
        if r[0] != ProtocolMsgNumbers.SSH_AGENT_IDENTITIES_ANSWER:
            raise SshAgentFailure('agent: non-identities answer received for identities list')
        return KeyList.from_blob(r[1:])

    def __contains__(self, public_key: CryptoPublicKey) -> bool:
        msg = PublicKeyMsg.from_public_key(public_key)
        return msg in self.list().keys


@functools.cache
def key_data_into_crypto_objects(key_data: bytes, passphrase: bytes | None) -> tuple[CryptoPrivateKey, CryptoPublicKey, str]:
    private_key = serialization.ssh.load_ssh_private_key(key_data, passphrase)
    public_key = private_key.public_key()
    fingerprint = PublicKeyMsg.from_public_key(public_key).fingerprint

    return private_key, public_key, fingerprint



#!/usr/bin/env python
from __future__ import annotations

import io
import os
import sys

import pexpect


env_vars = {
    'ANSIBLE_NOCOLOR': 'True',
    'ANSIBLE_RETRY_FILES_ENABLED': 'False',
}

env = os.environ.copy()
env.update(env_vars)

with io.BytesIO() as logfile:
    debugger_test_test = pexpect.spawn(
        'ansible-playbook',
        args=['test_run_once_playbook.yml'] + sys.argv[1:],
        timeout=10,
        env=env
    )

    debugger_test_test.logfile = logfile

    debugger_test_test.expect_exact('TASK: Task 1 (debug)> ')
    debugger_test_test.send('task.args["that"] = []\r')
    debugger_test_test.expect_exact('TASK: Task 1 (debug)> ')
    debugger_test_test.send('r\r')
    debugger_test_test.expect(pexpect.EOF)
    debugger_test_test.close()

    assert str(logfile.getvalue()).count('Task 2 executed') == 2



from __future__ import annotations

import dataclasses
import typing as t

from collections import abc as c

from ansible import constants
from ansible._internal._templating import _engine
from ansible._internal._templating._chain_templar import ChainTemplar
from ansible.errors import AnsibleError
from ansible.module_utils._internal._ambient_context import AmbientContextBase
from ansible.module_utils.datatag import native_type_name
from ansible.parsing import vault as _vault
from ansible.utils.display import Display

if t.TYPE_CHECKING:
    from ansible.playbook.task import Task


@dataclasses.dataclass
class TaskContext(AmbientContextBase):
    """Ambient context that wraps task execution on workers. It provides access to the currently executing task."""

    task: Task


TaskArgsFinalizerCallback = t.Callable[[str, t.Any, _engine.TemplateEngine, t.Any], t.Any]
"""Type alias for the shape of the `ActionBase.finalize_task_arg` method."""


class TaskArgsChainTemplar(ChainTemplar):
    """
    A ChainTemplar that carries a user-provided context object, optionally provided by `ActionBase.get_finalize_task_args_context`.
    TaskArgsFinalizer provides the context to each `ActionBase.finalize_task_arg` call to allow for more complex/stateful customization.
    """

    def __init__(self, *sources: c.Mapping, templar: _engine.TemplateEngine, callback: TaskArgsFinalizerCallback, context: t.Any) -> None:
        super().__init__(*sources, templar=templar)

        self.callback = callback
        self.context = context

    def template(self, key: t.Any, value: t.Any) -> t.Any:
        return self.callback(key, value, self.templar, self.context)


class TaskArgsFinalizer:
    """Invoked during task args finalization; allows actions to override default arg processing (e.g., templating)."""

    def __init__(self, *args: c.Mapping[str, t.Any] | str | None, templar: _engine.TemplateEngine) -> None:
        self._args_layers = [arg for arg in args if arg is not None]
        self._templar = templar

    def finalize(self, callback: TaskArgsFinalizerCallback, context: t.Any) -> dict[str, t.Any]:
        resolved_layers: list[c.Mapping[str, t.Any]] = []

        for layer in self._args_layers:
            if isinstance(layer, (str, _vault.EncryptedString)):  # EncryptedString can hide a template
                if constants.config.get_config_value('INJECT_FACTS_AS_VARS'):
                    Display().warning(
                        "Using a template for task args is unsafe in some situations "
                        "(see https://docs.ansible.com/ansible/devel/reference_appendices/faq.html#argsplat-unsafe).",
                        obj=layer,
                    )

                resolved_layer = self._templar.resolve_to_container(layer, options=_engine.TemplateOptions(value_for_omit={}))
            else:
                resolved_layer = layer

            if not isinstance(resolved_layer, dict):
                raise AnsibleError(f'Task args must resolve to a {native_type_name(dict)!r} not {native_type_name(resolved_layer)!r}.', obj=layer)

            resolved_layers.append(resolved_layer)

        ct = TaskArgsChainTemplar(*reversed(resolved_layers), templar=self._templar, callback=callback, context=context)

        return ct.as_dict()



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


class ModuleDocFragment(object):
    DOCUMENTATION = r"""
options:
    testcol2option:
        description:
            - A plugin option taken from testcol2
        type: str
        version_added: 1.0.0
        ini:
            - key: foo
              section: bar
              version_added: 1.1.0
              deprecated:
                alternative: none
                why: Test deprecation
                version: '3.0.0'
        env:
            - name: FOO_BAR
              version_added: 1.2.0
              deprecated:
                alternative: none
                why: Test deprecation
                removed_at_date: 2020-01-31
        vars:
            - name: foobar
              version_added: 1.3.0
              deprecated:
                alternative: none
                why: Test deprecation
                removed_at_date: 2040-12-31
    testcol2depr:
        description:
            - A plugin option taken from testcol2 that is deprecated
        type: str
        deprecated:
            alternative: none
            why: Test option deprecation
            version: '2.0.0'
"""



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os

import unittest
from unittest.mock import MagicMock, patch
from ansible.inventory.manager import InventoryManager
from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.playbook.play import Play


from units.mock.loader import DictDataLoader
from units.mock.path import mock_unfrackpath_noop

from ansible.vars.manager import VariableManager


def _tag_ignore_magicmock():
    orig_tag = AnsibleTagHelper.tag
    return patch.object(AnsibleTagHelper, 'tag', lambda value, *args, **kwargs: value if isinstance(value, MagicMock) else orig_tag(value, *args, **kwargs))


class TestVariableManager(unittest.TestCase):

    def test_basic_manager(self):
        fake_loader = DictDataLoader({})

        mock_inventory = InventoryManager(loader=fake_loader)
        v = VariableManager(loader=fake_loader, inventory=mock_inventory)
        variables = v.get_vars(use_cache=False)

        # Check var manager expected values,  never check: ['omit', 'vars']
        # FIXME:  add the following ['ansible_version', 'ansible_playbook_python', 'groups']
        for varname, value in (('playbook_dir', os.path.abspath('.')), ):
            self.assertEqual(variables[varname], value)

    def test_variable_manager_extra_vars(self):
        fake_loader = DictDataLoader({})

        extra_vars = dict(a=1, b=2, c=3)
        mock_inventory = InventoryManager(loader=fake_loader)
        v = VariableManager(loader=fake_loader, inventory=mock_inventory)

        # override internal extra_vars loading
        v._extra_vars = extra_vars

        myvars = v.get_vars(use_cache=False)
        for key, val in extra_vars.items():
            self.assertEqual(myvars.get(key), val)

    def test_variable_manager_options_vars(self):
        fake_loader = DictDataLoader({})

        options_vars = dict(a=1, b=2, c=3)
        mock_inventory = InventoryManager(loader=fake_loader)
        v = VariableManager(loader=fake_loader, inventory=mock_inventory)

        # override internal options_vars loading
        v._extra_vars = options_vars

        myvars = v.get_vars(use_cache=False)
        for key, val in options_vars.items():
            self.assertEqual(myvars.get(key), val)

    def test_variable_manager_play_vars(self):
        fake_loader = DictDataLoader({})

        mock_play = MagicMock()
        mock_play.get_vars.return_value = dict(foo="bar")
        mock_play.get_roles.return_value = []
        mock_play.get_vars_files.return_value = []

        mock_inventory = MagicMock()

        v = VariableManager(loader=fake_loader, inventory=mock_inventory)

        with _tag_ignore_magicmock():
            self.assertEqual(v.get_vars(play=mock_play, use_cache=False).get("foo"), "bar")

    def test_variable_manager_play_vars_files(self):
        fake_loader = DictDataLoader({
            __file__: """
               foo: bar
            """
        })

        mock_play = MagicMock()
        mock_play.get_vars.return_value = dict()
        mock_play.get_roles.return_value = []
        mock_play.get_vars_files.return_value = [__file__]

        mock_inventory = MagicMock()
        v = VariableManager(inventory=mock_inventory, loader=fake_loader)
        with _tag_ignore_magicmock():
            self.assertEqual(v.get_vars(play=mock_play, use_cache=False).get("foo"), "bar")

    @patch('ansible.playbook.role.definition.unfrackpath', mock_unfrackpath_noop)
    def test_variable_manager_role_vars_dependencies(self):
        """
        Tests vars from role dependencies with duplicate dependencies.
        """
        fake_loader = DictDataLoader({
            # role common-role
            '/etc/ansible/roles/common-role/tasks/main.yml': """
            - debug: msg="{{role_var}}"
            """,
            # We do not need allow_duplicates: yes for this role
            # because eliminating duplicates is done by the execution
            # strategy, which we do not test here.

            # role role1
            '/etc/ansible/roles/role1/vars/main.yml': """
            role_var: "role_var_from_role1"
            """,
            '/etc/ansible/roles/role1/meta/main.yml': """
            dependencies:
              - { role: common-role }
            """,

            # role role2
            '/etc/ansible/roles/role2/vars/main.yml': """
            role_var: "role_var_from_role2"
            """,
            '/etc/ansible/roles/role2/meta/main.yml': """
            dependencies:
              - { role: common-role }
            """,
        })

        mock_inventory = InventoryManager(loader=fake_loader)

        with _tag_ignore_magicmock():
            v = VariableManager(loader=fake_loader, inventory=mock_inventory)

            play1 = Play.load(dict(
                hosts=['all'],
                roles=['role1', 'role2'],
            ), loader=fake_loader, variable_manager=v)

            # The task defined by common-role exists twice because role1
            # and role2 depend on common-role.  Check that the tasks see
            # different values of role_var.
            blocks = play1.compile()
            task = blocks[1].block[0]
            res = v.get_vars(play=play1, task=task)
            self.assertEqual(res['role_var'], 'role_var_from_role1')

            task = blocks[2].block[0]
            res = v.get_vars(play=play1, task=task)
            self.assertEqual(res['role_var'], 'role_var_from_role2')



#!/usr/bin/python
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
module: invalid_argument_spec_incorrect_context
short_description: Invalid argument spec incorrect context schema test module
description: Invalid argument spec incorrect context schema test module
author:
  - Ansible Core Team
options:
  foo:
    description: foo
    type: str
"""

EXAMPLES = """#"""
RETURN = """"""

from ansible.module_utils.basic import AnsibleModule


def main():
    AnsibleModule(
        argument_spec=dict(
            foo=dict(
                type='str',
                context='bar',
            ),
        ),
    )


if __name__ == '__main__':
    main()



# Copyright: (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import re

from ansible import constants as C
from ansible.errors import AnsibleError
from ansible.utils.display import Display
from ansible.utils.plugin_docs import get_versioned_doclink

_FALLBACK_INTERPRETER = '/usr/bin/python3'

display = Display()
foundre = re.compile(r'FOUND(.*)ENDFOUND', flags=re.DOTALL)


class InterpreterDiscoveryRequiredError(Exception):
    def __init__(self, message, interpreter_name, discovery_mode):
        super(InterpreterDiscoveryRequiredError, self).__init__(message)
        self.interpreter_name = interpreter_name
        self.discovery_mode = discovery_mode


def discover_interpreter(action, interpreter_name, discovery_mode, task_vars):
    """Probe the target host for a Python interpreter from the `INTERPRETER_PYTHON_FALLBACK` list, returning the first found or `/usr/bin/python3` if none."""
    host = task_vars.get('inventory_hostname', 'unknown')
    res = None
    found_interpreters = [_FALLBACK_INTERPRETER]  # fallback value
    is_silent = discovery_mode.endswith('_silent')

    if discovery_mode.startswith('auto_legacy'):
        display.deprecated(
            msg=f"The '{discovery_mode}' option for 'INTERPRETER_PYTHON' now has the same effect as 'auto'.",
            version='2.21',
        )

    try:
        bootstrap_python_list = C.config.get_config_value('INTERPRETER_PYTHON_FALLBACK', variables=task_vars)

        display.vvv(msg=f"Attempting {interpreter_name} interpreter discovery.", host=host)

        # not all command -v impls accept a list of commands, so we have to call it once per python
        command_list = ["command -v '%s'" % py for py in bootstrap_python_list]
        shell_bootstrap = "echo FOUND; {0}; echo ENDFOUND".format('; '.join(command_list))

        # FUTURE: in most cases we probably don't want to use become, but maybe sometimes we do?
        res = action._low_level_execute_command(shell_bootstrap, sudoable=False)

        raw_stdout = res.get('stdout', u'')

        match = foundre.match(raw_stdout)

        if not match:
            display.debug(u'raw interpreter discovery output: {0}'.format(raw_stdout), host=host)
            raise ValueError('unexpected output from Python interpreter discovery')

        found_interpreters = [interp.strip() for interp in match.groups()[0].splitlines() if interp.startswith('/')]

        display.debug(u"found interpreters: {0}".format(found_interpreters), host=host)

        if not found_interpreters:
            if not is_silent:
                display.warning(msg=f'No python interpreters found for host {host!r} (tried {bootstrap_python_list!r}).')

            # this is lame, but returning None or throwing an exception is uglier
            return _FALLBACK_INTERPRETER
    except AnsibleError:
        raise
    except Exception as ex:
        if not is_silent:
            display.error_as_warning(msg=f'Unhandled error in Python interpreter discovery for host {host!r}.', exception=ex)

            if res and res.get('stderr'):  # the current ssh plugin implementation always has stderr, making coverage of the false case difficult
                display.vvv(msg=f"Interpreter discovery remote stderr:\n{res.get('stderr')}", host=host)

    if not is_silent:
        display.warning(
            msg=(
                f"Host {host!r} is using the discovered Python interpreter at {found_interpreters[0]!r}, "
                "but future installation of another Python interpreter could cause a different interpreter to be discovered."
            ),
            help_text=f"See {get_versioned_doclink('reference_appendices/interpreter_discovery.html')} for more information.",
        )

    return found_interpreters[0]



# -*- coding: utf-8 -*-
# Copyright:
#   (c) 2023 Ansible Project
# License: GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from ansible.module_utils.testing import patch_module_args
from ansible.modules import uri


class TestUri:

    def test_main_no_args(self):
        """Module must fail if called with no args."""
        with pytest.raises(SystemExit), \
             patch_module_args():
            uri.main()

    def test_main_no_force(self, mocker):
        """The "force" parameter to fetch_url() must be absent or false when the module is called without "force"."""
        resp = MagicMock()
        resp.headers.get_content_type.return_value = "text/html"
        info = {"url": "http://example.com/", "status": 200}
        with patch.object(uri, "fetch_url", return_value=(resp, info)) as fetch_url, \
             patch_module_args({"url": "http://example.com/"}):
            with pytest.raises(SystemExit):
                uri.main()
            fetch_url.assert_called_once()
            assert not fetch_url.call_args[1].get("force")

    def test_main_force(self):
        """The "force" parameter to fetch_url() must be true when the module is called with "force"."""
        resp = MagicMock()
        resp.headers.get_content_type.return_value = "text/html"
        info = {"url": "http://example.com/", "status": 200}
        with patch.object(uri, "fetch_url", return_value=(resp, info)) as fetch_url, \
             patch_module_args({"url": "http://example.com/", "force": True}):
            with pytest.raises(SystemExit):
                uri.main()
            fetch_url.assert_called_once()
            assert fetch_url.call_args[1].get("force")



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule


def main() -> None:
    m = AnsibleModule({})
    m.warn("Hello\r\nNew\rAnsible\nWorld")
    m.exit_json()


if __name__ == '__main__':
    main()



#!/usr/bin/python
from __future__ import annotations

import json


def main():
    print(json.dumps(dict(changed=False, failed=True, msg='this collection should be masked by testcoll in the user content root')))


if __name__ == '__main__':
    main()



# (c) 2012-2014, Chris Meyers <chris.meyers.fsu@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import json
import re
import textwrap
import types

import unittest
from unittest.mock import MagicMock

from ansible.executor.task_result import CallbackTaskResult
from ansible.inventory.host import Host
from ansible.plugins.callback import CallbackBase


mock_task = MagicMock()
mock_task.delegate_to = None


class TestCallback(unittest.TestCase):
    # FIXME: This doesn't really test anything...
    def test_init(self):
        CallbackBase()

    def test_display(self):
        display_mock = MagicMock()
        display_mock.verbosity = 0
        cb = CallbackBase(display=display_mock)
        self.assertIs(cb._display, display_mock)

    def test_display_verbose(self):
        display_mock = MagicMock()
        display_mock.verbosity = 5
        cb = CallbackBase(display=display_mock)
        self.assertIs(cb._display, display_mock)

    def test_host_label(self):
        result = CallbackTaskResult(host=Host('host1'), task=mock_task, return_data={}, task_fields={})

        self.assertEqual(CallbackBase.host_label(result), 'host1')

    def test_host_label_delegated(self):
        mock_task.delegate_to = 'host2'
        result = CallbackTaskResult(
            host=Host('host1'),
            task=mock_task,
            return_data={'_ansible_delegated_vars': {'ansible_host': 'host2'}},
            task_fields={},
        )
        self.assertEqual(CallbackBase.host_label(result), 'host1 -> host2')

    # TODO: import callback module so we can patch callback.cli/callback.C


class TestCallbackResults(unittest.TestCase):

    def test_get_item_label(self):
        cb = CallbackBase()
        results = {'item': 'some_item'}
        res = cb._get_item_label(results)
        self.assertEqual(res, 'some_item')

    def test_get_item_label_no_log(self):
        cb = CallbackBase()
        results = {'item': 'some_item', '_ansible_no_log': True}
        res = cb._get_item_label(results)
        self.assertEqual(res, "(censored due to no_log)")

        results = {'item': 'some_item', '_ansible_no_log': False}
        res = cb._get_item_label(results)
        self.assertEqual(res, "some_item")

    def test_clean_results_debug_task(self):
        cb = CallbackBase()
        result = {'item': 'some_item',
                  'invocation': 'foo --bar whatever [some_json]',
                  'a': 'a single a in result note letter a is in invocation',
                  'b': 'a single b in result note letter b is not in invocation',
                  'changed': True}

        cb._clean_results(result, 'debug')

        # See https://github.com/ansible/ansible/issues/33723
        self.assertTrue('a' in result)
        self.assertTrue('b' in result)
        self.assertFalse('invocation' in result)
        self.assertFalse('changed' in result)

    def test_clean_results_debug_task_no_invocation(self):
        cb = CallbackBase()
        result = {'item': 'some_item',
                  'a': 'a single a in result note letter a is in invocation',
                  'b': 'a single b in result note letter b is not in invocation',
                  'changed': True}

        cb._clean_results(result, 'debug')
        self.assertTrue('a' in result)
        self.assertTrue('b' in result)
        self.assertFalse('changed' in result)
        self.assertFalse('invocation' in result)

    def test_clean_results_debug_task_empty_results(self):
        cb = CallbackBase()
        result = {}
        cb._clean_results(result, 'debug')
        self.assertFalse('invocation' in result)
        self.assertEqual(len(result), 0)

    def test_clean_results(self):
        cb = CallbackBase()
        result = {'item': 'some_item',
                  'invocation': 'foo --bar whatever [some_json]',
                  'a': 'a single a in result note letter a is in invocation',
                  'b': 'a single b in result note letter b is not in invocation',
                  'changed': True}

        expected_result = result.copy()
        cb._clean_results(result, 'ebug')
        self.assertEqual(result, expected_result)


class TestCallbackDumpResults(object):
    def test_internal_keys(self):
        cb = CallbackBase()
        result = {'item': 'some_item',
                  '_ansible_some_var': 'SENTINEL',
                  'testing_ansible_out': 'should_be_left_in LEFTIN',
                  'invocation': 'foo --bar whatever [some_json]',
                  'some_dict_key': {'a_sub_dict_for_key': 'baz'},
                  'bad_dict_key': {'_ansible_internal_blah': 'SENTINEL'},
                  'changed': True}
        json_out = cb._dump_results(result)
        assert '"_ansible_' not in json_out
        assert 'SENTINEL' not in json_out
        assert 'LEFTIN' in json_out

    def test_exception(self):
        cb = CallbackBase()
        result = {'item': 'some_item LEFTIN',
                  'exception': ['frame1', 'SENTINEL']}
        json_out = cb._dump_results(result)
        assert 'SENTINEL' not in json_out
        assert 'exception' not in json_out
        assert 'LEFTIN' in json_out

    def test_verbose(self):
        cb = CallbackBase()
        result = {'item': 'some_item LEFTIN',
                  '_ansible_verbose_always': 'chicane'}
        json_out = cb._dump_results(result)
        assert 'SENTINEL' not in json_out
        assert 'LEFTIN' in json_out

    def test_diff(self):
        cb = CallbackBase()
        result = {'item': 'some_item LEFTIN',
                  'diff': ['remove stuff', 'added LEFTIN'],
                  '_ansible_verbose_always': 'chicane'}
        json_out = cb._dump_results(result)
        assert 'SENTINEL' not in json_out
        assert 'LEFTIN' in json_out

    def test_mixed_keys(self):
        cb = CallbackBase()
        result = {3: 'pi',
                  'tau': 6}
        json_out = cb._dump_results(result)
        round_trip_result = json.loads(json_out)
        assert len(round_trip_result) == 2
        assert '3' in round_trip_result
        assert 'tau' in round_trip_result
        assert round_trip_result['3'] == 'pi'
        assert round_trip_result['tau'] == 6


class TestCallbackDiff(unittest.TestCase):

    def setUp(self):
        self.cb = CallbackBase()

    def _strip_color(self, s):
        return re.sub('\033\\[[^m]*m', '', s)

    def test_difflist(self):
        # TODO: split into smaller tests?
        difflist = [{'before': u'preface\nThe Before String\npostscript',
                     'after': u'preface\nThe After String\npostscript',
                     'before_header': u'just before',
                     'after_header': u'just after'
                     },
                    {'before': u'preface\nThe Before String\npostscript',
                     'after': u'preface\nThe After String\npostscript',
                     },
                    {'src_binary': 'chicane'},
                    {'dst_binary': 'chicanery'},
                    {'dst_larger': 1},
                    {'src_larger': 2},
                    {'prepared': u'what does prepared do?'},
                    {'before_header': u'just before'},
                    {'after_header': u'just after'}]

        res = self.cb._get_diff(difflist)

        self.assertIn(u'Before String', res)
        self.assertIn(u'After String', res)
        self.assertIn(u'just before', res)
        self.assertIn(u'just after', res)

    def test_simple_diff(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': 'one\ntwo\nthree\n',
                'after': 'one\nthree\nfour\n',
            })),
            textwrap.dedent("""\
                --- before: somefile.txt
                +++ after: generated from template somefile.j2
                @@ -1,3 +1,3 @@
                 one
                -two
                 three
                +four

            """))

    def test_new_file(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': '',
                'after': 'one\ntwo\nthree\n',
            })),
            textwrap.dedent("""\
                --- before: somefile.txt
                +++ after: generated from template somefile.j2
                @@ -0,0 +1,3 @@
                +one
                +two
                +three

            """))

    def test_clear_file(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': 'one\ntwo\nthree\n',
                'after': '',
            })),
            textwrap.dedent("""\
                --- before: somefile.txt
                +++ after: generated from template somefile.j2
                @@ -1,3 +0,0 @@
                -one
                -two
                -three

            """))

    def test_no_trailing_newline_before(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': 'one\ntwo\nthree',
                'after': 'one\ntwo\nthree\n',
            })),
            textwrap.dedent("""\
                --- before: somefile.txt
                +++ after: generated from template somefile.j2
                @@ -1,3 +1,3 @@
                 one
                 two
                -three
                \\ No newline at end of file
                +three

            """))

    def test_no_trailing_newline_after(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': 'one\ntwo\nthree\n',
                'after': 'one\ntwo\nthree',
            })),
            textwrap.dedent("""\
                --- before: somefile.txt
                +++ after: generated from template somefile.j2
                @@ -1,3 +1,3 @@
                 one
                 two
                -three
                +three
                \\ No newline at end of file

            """))

    def test_no_trailing_newline_both(self):
        self.assertMultiLineEqual(
            self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': 'one\ntwo\nthree',
                'after': 'one\ntwo\nthree',
            }),
            '')

    def test_no_trailing_newline_both_with_some_changes(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before_header': 'somefile.txt',
                'after_header': 'generated from template somefile.j2',
                'before': 'one\ntwo\nthree',
                'after': 'one\nfive\nthree',
            })),
            textwrap.dedent("""\
                --- before: somefile.txt
                +++ after: generated from template somefile.j2
                @@ -1,3 +1,3 @@
                 one
                -two
                +five
                 three
                \\ No newline at end of file

            """))

    def test_diff_dicts(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before': dict(one=1, two=2, three=3),
                'after': dict(one=1, three=3, four=4),
            })),
            textwrap.dedent("""\
                --- before
                +++ after
                @@ -1,5 +1,5 @@
                 {
                +    "four": 4,
                     "one": 1,
                -    "three": 3,
                -    "two": 2
                +    "three": 3
                 }

            """))

    def test_diff_before_none(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before': None,
                'after': 'one line\n',
            })),
            textwrap.dedent("""\
                --- before
                +++ after
                @@ -0,0 +1 @@
                +one line

            """))

    def test_diff_after_none(self):
        self.assertMultiLineEqual(
            self._strip_color(self.cb._get_diff({
                'before': 'one line\n',
                'after': None,
            })),
            textwrap.dedent("""\
                --- before
                +++ after
                @@ -1 +0,0 @@
                -one line

            """))


class TestCallbackOnMethods(unittest.TestCase):
    def _find_on_methods(self, callback):
        cb_dir = dir(callback)
        method_names = [x for x in cb_dir if '_on_' in x]
        methods = [getattr(callback, mn) for mn in method_names]
        return methods

    def test_are_methods(self):
        cb = CallbackBase()
        for method in self._find_on_methods(cb):
            self.assertIsInstance(method, types.MethodType)

    def test_on_any(self):
        cb = CallbackBase()
        cb.v2_on_any('whatever', some_keyword='blippy')
        cb.on_any('whatever', some_keyword='blippy')


def test_v2_v1_method_map() -> None:
    """Ensure that all v2 callback methods appear in the method map."""
    expected_names = [name for name in dir(CallbackBase) if name.startswith('v2_')]
    mapped_names = {method.__name__ for method in CallbackBase._v2_v1_method_map}

    missing = [name for name in expected_names if name not in mapped_names]

    assert not missing



"""Handling of `Marker` values."""

from __future__ import annotations

import abc
import contextlib
import dataclasses
import itertools
import typing as t

from ansible.utils.display import Display

from ._jinja_common import Marker


class MarkerBehavior(metaclass=abc.ABCMeta):
    """Base class to support custom handling of `Marker` values encountered during concatenation or finalization."""

    @abc.abstractmethod
    def handle_marker(self, value: Marker) -> t.Any:
        """Handle the given `Marker` value."""


class FailingMarkerBehavior(MarkerBehavior):
    """
    The default behavior when encountering a `Marker` value during concatenation or finalization.
    This always raises the template-internal `MarkerError` exception.
    """

    def handle_marker(self, value: Marker) -> t.Any:
        value.trip()


# FAIL_ON_MARKER_BEHAVIOR
# _DETONATE_MARKER_BEHAVIOR - internal singleton since it's the default and nobody should need to reference it, or make it an actual singleton
FAIL_ON_UNDEFINED: t.Final = FailingMarkerBehavior()  # no sense in making many instances...


@dataclasses.dataclass(kw_only=True, slots=True, frozen=True)
class _MarkerTracker:
    """A numbered occurrence of a `Marker` value for later conversion to a warning."""

    number: int
    value: Marker


class ReplacingMarkerBehavior(MarkerBehavior):
    """All `Marker` values are replaced with a numbered string placeholder and the message from the value."""

    def __init__(self) -> None:
        self._trackers: list[_MarkerTracker] = []

    def record_marker(self, value: Marker) -> t.Any:
        """Assign a sequence number to the given value and record it for later generation of warnings."""
        number = len(self._trackers) + 1

        self._trackers.append(_MarkerTracker(number=number, value=value))

        return number

    def emit_warnings(self) -> None:
        """Emit warning messages caused by Marker values, aggregated by unique template."""

        display = Display()
        grouped_templates = itertools.groupby(self._trackers, key=lambda tracker: tracker.value._marker_template_source)

        for template, items in grouped_templates:
            item_list = list(items)

            msg = f'Encountered {len(item_list)} template error{"s" if len(item_list) > 1 else ""}.'

            for item in item_list:
                msg += f'\nerror {item.number} - {item.value._as_message()}'

            display.warning(msg=msg, obj=template)

    @classmethod
    @contextlib.contextmanager
    def warning_context(cls) -> t.Generator[t.Self, None, None]:
        """Collect warnings for `Marker` values and emit warnings when the context exits."""
        instance = cls()

        try:
            yield instance
        finally:
            instance.emit_warnings()

    def handle_marker(self, value: Marker) -> t.Any:
        number = self.record_marker(value)

        return f"<< error {number} - {value._as_message()} >>"


class RoutingMarkerBehavior(MarkerBehavior):
    """Routes instances of Marker (by type reference) to another MarkerBehavior, defaulting to FailingMarkerBehavior."""

    def __init__(self, dispatch_table: dict[type[Marker], MarkerBehavior]) -> None:
        self._dispatch_table = dispatch_table

    def handle_marker(self, value: Marker) -> t.Any:
        behavior = self._dispatch_table.get(type(value), FAIL_ON_UNDEFINED)

        return behavior.handle_marker(value)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import unittest
from unittest.mock import patch

from ansible.module_utils import basic
from ansible.modules.service_facts import AIXScanService


# AIX # lssrc -a
LSSRC_OUTPUT = """
Subsystem         Group            PID          Status
 sendmail         mail             5243302      active
 syslogd          ras              5636528      active
 portmap          portmap          5177768      active
 snmpd            tcpip            5308844      active
 hostmibd         tcpip            5374380      active
 snmpmibd         tcpip            5439918      active
 aixmibd          tcpip            5505456      active
 nimsh            nimclient        5571004      active
 aso                               6029758      active
 biod             nfs              6357464      active
 nfsd             nfs              5701906      active
 rpc.mountd       nfs              6488534      active
 rpc.statd        nfs              7209216      active
 rpc.lockd        nfs              7274988      active
 qdaemon          spooler          6816222      active
 writesrv         spooler          6685150      active
 clcomd           caa              7471600      active
 sshd             ssh              7602674      active
 pfcdaemon                         7012860      active
 ctrmc            rsct             6947312      active
 IBM.HostRM       rsct_rm          14418376     active
 IBM.ConfigRM     rsct_rm          6160674      active
 IBM.DRM          rsct_rm          14680550     active
 IBM.MgmtDomainRM rsct_rm          14090676     active
 IBM.ServiceRM    rsct_rm          13828542     active
 cthats           cthats           13959668     active
 cthags           cthags           14025054     active
 IBM.StorageRM    rsct_rm          12255706     active
 inetd            tcpip            12517828     active
 lpd              spooler                       inoperative
 keyserv          keyserv                       inoperative
 ypbind           yp                            inoperative
 gsclvmd                                        inoperative
 cdromd                                         inoperative
 ndpd-host        tcpip                         inoperative
 ndpd-router      tcpip                         inoperative
 netcd            netcd                         inoperative
 tftpd            tcpip                         inoperative
 routed           tcpip                         inoperative
 mrouted          tcpip                         inoperative
 rsvpd            qos                           inoperative
 policyd          qos                           inoperative
 timed            tcpip                         inoperative
 iptrace          tcpip                         inoperative
 dpid2            tcpip                         inoperative
 rwhod            tcpip                         inoperative
 pxed             tcpip                         inoperative
 binld            tcpip                         inoperative
 xntpd            tcpip                         inoperative
 gated            tcpip                         inoperative
 dhcpcd           tcpip                         inoperative
 dhcpcd6          tcpip                         inoperative
 dhcpsd           tcpip                         inoperative
 dhcpsdv6         tcpip                         inoperative
 dhcprd           tcpip                         inoperative
 dfpd             tcpip                         inoperative
 named            tcpip                         inoperative
 automountd       autofs                        inoperative
 nfsrgyd          nfs                           inoperative
 gssd             nfs                           inoperative
 cpsd             ike                           inoperative
 tmd              ike                           inoperative
 isakmpd                                        inoperative
 ikev2d                                         inoperative
 iked             ike                           inoperative
 clconfd          caa                           inoperative
 ksys_vmmon                                     inoperative
 nimhttp                                        inoperative
 IBM.SRVPROXY     ibmsrv                        inoperative
 ctcas            rsct                          inoperative
 IBM.ERRM         rsct_rm                       inoperative
 IBM.AuditRM      rsct_rm                       inoperative
 isnstgtd         isnstgtd                      inoperative
 IBM.LPRM         rsct_rm                       inoperative
 cthagsglsm       cthags                        inoperative
"""


class TestAIXScanService(unittest.TestCase):

    def setUp(self):
        self.mock1 = patch.object(basic.AnsibleModule, 'get_bin_path', return_value='/usr/sbin/lssrc')
        self.mock1.start()
        self.addCleanup(self.mock1.stop)
        self.mock2 = patch.object(basic.AnsibleModule, 'run_command', return_value=(0, LSSRC_OUTPUT, ''))
        self.mock2.start()
        self.addCleanup(self.mock2.stop)
        self.mock3 = patch('platform.system', return_value='AIX')
        self.mock3.start()
        self.addCleanup(self.mock3.stop)

    def test_gather_services(self):
        svcmod = AIXScanService(basic.AnsibleModule)
        result = svcmod.gather_services()

        self.assertIsInstance(result, dict)

        self.assertIn('IBM.HostRM', result)
        self.assertEqual(result['IBM.HostRM'], {
            'name': 'IBM.HostRM',
            'source': 'src',
            'state': 'running',
        })
        self.assertIn('IBM.AuditRM', result)
        self.assertEqual(result['IBM.AuditRM'], {
            'name': 'IBM.AuditRM',
            'source': 'src',
            'state': 'stopped',
        })
